import logging
from logging.handlers import RotatingFileHandler
from concurrent_log_handler import ConcurrentRotatingFileHandler
import os
import inspect
from pathlib import Path
import datetime
import abc
from rich.logging import RichHandler
from rich.console import Console
import traceback


def print(*args, sep=" ", end="\n"):
    logger = logging.getLogger("debug")
    str_list = [str(i) for i in args]
    msg = sep.join(str_list) + end
    logger.debug(msg)


def show_traceback(rich_print=True):
    if rich_print:
        console = Console()
        console.print_exception()
    return traceback.format_exc()


class BaseLogger(metaclass=abc.ABCMeta):
    _saved_loggers = {}

    @abc.abstractmethod
    def get_logger(self, name, **kwargs):
        pass

    @abc.abstractmethod
    def debug(self, *msg, sep=" ", **kwargs):
        pass

    @abc.abstractmethod
    def info(self, *msg, sep=" ", **kwargs):
        pass

    @abc.abstractmethod
    def warning(self, *msg, sep=" ", **kwargs):
        pass

    @abc.abstractmethod
    def error(self, *msg, sep=" ", **kwargs):
        pass

    @staticmethod
    def _get_format_msg(currentframe, msg: tuple, level, sep=" "):
        filename = os.path.basename(currentframe.f_back.f_code.co_filename)
        lineno = currentframe.f_back.f_lineno
        msg_list = [str(i) for i in msg]
        msg = sep.join(msg_list)
        msg = f"[{filename}] [line:{(lineno): ^4}] {level: ^7} >>> " + msg
        return msg

    @staticmethod
    def _get_format_logger(name, log_abs_path, level=logging.INFO, multi_process=True, stream=True):
        default_formats = {
            "color_format": "%(log_color)s%(asctime)s-%(message)s",
            "log_format": "%(asctime)s-%(message)s",
        }

        log_path = Path(log_abs_path).absolute()
        log_dir = log_path.parent
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        logger = logging.Logger(name, level=level)

        file_formatter = logging.Formatter(
            default_formats["log_format"],
            datefmt="%y/%m/%d %H:%M:%S",
        )

        if multi_process:
            rotating_file_handler = ConcurrentRotatingFileHandler
        else:
            rotating_file_handler = RotatingFileHandler

        file_handler = rotating_file_handler(
            filename=log_path,
            maxBytes=10 * 1024 * 1024,
            backupCount=10,
            encoding="utf-8",
        )
        file_handler.setFormatter(file_formatter)
        logger.addHandler(file_handler)

        if stream:
            logger.addHandler(RichHandler())
        # logger.setLevel(level)
        return logger


class SimpleLogger(BaseLogger):

    def __init__(
            self,
            name="name",
            log_dir="./logs",
            print_stream=True,
            level=logging.DEBUG,
            multi_process=False,
            tz_is_china=True,
    ):
        if tz_is_china:
            logging.Formatter.converter = lambda sec, what: (
                    datetime.datetime.now(tz=datetime.timezone.utc)
                    + datetime.timedelta(hours=8)
            ).timetuple()
        log_path = Path(log_dir).joinpath(name + '.log')

        self._logger = self._get_format_logger(
            f"debug-{name}", log_path, level=level, stream=print_stream, multi_process=multi_process
        )
        self._saved_loggers[name] = self

    @classmethod
    def get_logger(cls,
                   name,
                   log_dir="./logs",
                   print_stream=True,
                   level=logging.DEBUG,
                   multi_process=False,
                   tz_is_china=True) -> "SimpleLogger":
        if name in cls._saved_loggers:
            return cls._saved_loggers[name]
        else:
            return cls(name,
                       log_dir=log_dir,
                       print_stream=print_stream,
                       level=level,
                       multi_process=multi_process,
                       tz_is_china=tz_is_china)

    def log(self, level, *msg, sep=" ", **kwargs):
        currentframe = inspect.currentframe()
        msg = self._get_format_msg(currentframe, msg, logging.getLevelName(level), sep=sep)
        self._logger.log(level, msg, **kwargs)

    def debug(self, *msg, sep=" ", **kwargs):
        currentframe = inspect.currentframe()
        msg = self._get_format_msg(currentframe, msg, "DEBUG", sep=sep)
        self._logger.debug(msg, **kwargs)

    def info(self, *msg, sep=" ", **kwargs):
        currentframe = inspect.currentframe()
        msg = self._get_format_msg(currentframe, msg, "INFO", sep=sep)
        self._logger.info(msg, **kwargs)

    def warning(self, *msg, sep=" ", **kwargs):
        currentframe = inspect.currentframe()
        msg = self._get_format_msg(currentframe, msg, "WARNING", sep=sep)
        self._logger.warning(msg, **kwargs)

    def error(self, *msg, sep=" ", **kwargs):
        currentframe = inspect.currentframe()
        msg = self._get_format_msg(currentframe, msg, "ERROR", sep=sep)
        self._logger.error(msg, **kwargs)


class Logger(BaseLogger):
    """
    Examples:
        >>> logger = Logger(name='train-log', log_dir='./log', print_debug=True)
        >>> logger.debug("hello", "list", [1, 2, 3, 4, 5])

        >>> logger2 = Logger.get_logger('train-log')
        >>> id(logger2) == id(logger)
        >>> True

    """

    def __init__(
            self,
            name="name",
            log_dir="./logs",
            debug_path="debug.log",
            info_path="info.log",
            warning_path="warn.log",
            error_path="error.log",
            multi_process=True,
            print_stream=True,
            print_debug=False,
            print_info=False,
            print_warning=False,
            print_error=False,
            single_mode=False,
            level=logging.DEBUG,
            tz_is_china=True,
    ):
        """
        Parameters
        ----------
            tz_is_china: bool
                time zone is China or not
        """
        if tz_is_china:
            logging.Formatter.converter = lambda sec, what: (
                    datetime.datetime.now(tz=datetime.timezone.utc)
                    + datetime.timedelta(hours=8)
            ).timetuple()
        if print_stream:
            if not any([print_debug, print_info, print_warning, print_error]):
                print_info, print_warning, print_error = True, True, True
            elif print_debug:
                print_info, print_warning, print_error = True, True, True
            elif print_info:
                print_error, print_warning = True, True
            elif print_warning:
                print_error = True

        debug_path = Path(log_dir).joinpath(debug_path)
        info_path = Path(log_dir).joinpath(info_path)
        warning_path = Path(log_dir).joinpath(warning_path)
        error_path = Path(log_dir).joinpath(error_path)

        self._debug_logger = self._get_format_logger(
            f"debug-{name}", debug_path, level=logging.DEBUG, stream=print_debug, multi_process=multi_process
        )
        self._info_logger = self._get_format_logger(
            f"info-{name}", info_path, level=logging.INFO, stream=print_info, multi_process=multi_process
        )
        self._warining_logger = self._get_format_logger(
            f"warning-{name}", warning_path, level=logging.WARNING, stream=print_warning, multi_process=multi_process
        )
        self._error_logger = self._get_format_logger(
            f"error-{name}", error_path, level=logging.ERROR, stream=print_error, multi_process=multi_process
        )
        self._single_mode = single_mode
        self._level = level
        self._saved_loggers[name] = self

    @classmethod
    def get_logger(cls, name, **kwargs) -> "Logger":
        if name in cls._saved_loggers:
            return cls._saved_loggers[name]
        else:
            return cls(name=name, **kwargs)

    def debug(self, *msg, sep=" ", **kwargs):
        currentframe = inspect.currentframe()
        msg = self._get_format_msg(currentframe, msg, "DEBUG", sep=sep)
        if self._level <= logging.DEBUG:
            self._debug_logger.debug(msg, **kwargs)

    def info(self, *msg, sep=" ", **kwargs):
        currentframe = inspect.currentframe()
        msg = self._get_format_msg(currentframe, msg, "INFO", sep=sep)
        if self._level <= logging.INFO:
            self._info_logger.info(msg, **kwargs)
            if not self._single_mode:
                self._debug_logger.info(msg, **kwargs)

    def warning(self, *msg, sep=" ", **kwargs):
        currentframe = inspect.currentframe()
        msg = self._get_format_msg(currentframe, msg, "WARNING", sep=sep)
        if self._level <= logging.WARNING:
            self._warining_logger.warning(msg, **kwargs)
            if not self._single_mode:
                self._debug_logger.warning(msg, **kwargs)
                self._info_logger.warning(msg, **kwargs)

    def error(self, *msg, sep=" ", **kwargs):
        currentframe = inspect.currentframe()
        msg = self._get_format_msg(currentframe, msg, "ERROR", sep=sep)
        if self._level <= logging.ERROR:
            self._error_logger.error(msg, **kwargs)
            if not self._single_mode:
                self._debug_logger.error(msg, **kwargs)
                self._info_logger.error(msg, **kwargs)
                self._warining_logger.error(msg, **kwargs)



import os
from fastapi import FastAPI
from fastapi.openapi.docs import (
    get_redoc_html,
    get_swagger_ui_html,
    get_swagger_ui_oauth2_redirect_html,
)
from fastapi.middleware.cors import CORSMiddleware
from fastapi import APIRouter
from fastapi.staticfiles import StaticFiles
from fastapi.responses import JSONResponse
import orjson


class ORJSONResponse(JSONResponse):
    media_type = "application/json"

    def render(self, content) -> bytes:
        return orjson.dumps(content, option=orjson.OPT_SERIALIZE_NUMPY)


def set_app_doc_static_file(app, favicon_url="icon.png"):
    """
    favicon_url :str: "icon.png" or "icon-yuanian.png"
    """
    root = os.path.dirname(os.path.realpath(__file__))
    app.mount("/static", StaticFiles(directory=f"{root}/static"), name="static")

    @app.get('/docs', include_in_schema=False)
    async def custom_swagger_ui_html():
        return get_swagger_ui_html(
            openapi_url=app.openapi_url,
            title=app.title + " - Swagger UI",
            oauth2_redirect_url=app.swagger_ui_oauth2_redirect_url,
            swagger_js_url="/static/swagger-ui-bundle.js",
            swagger_css_url="/static/swagger-ui.css",
            swagger_favicon_url=f"/static/{favicon_url}",
        )

    @app.get(app.swagger_ui_oauth2_redirect_url, include_in_schema=False)
    async def swagger_ui_redirect():
        return get_swagger_ui_oauth2_redirect_html()

    @app.get("/redoc", include_in_schema=False)
    async def redoc_html():
        return get_redoc_html(
            openapi_url=app.openapi_url,
            title=app.title + " - ReDoc",
            redoc_js_url="/static/redoc.standalone.js",
            redoc_favicon_url=f"/static/{favicon_url}",
        )

    @app.get("/8dbd28457ff989b4568a.worker.js.map", include_in_schema=False)
    async def redoc_worker_html():
        return get_redoc_html(
            openapi_url=app.openapi_url,
            title=app.title + " - worker.js.map",
            redoc_js_url="/static/8dbd28457ff989b4568a.worker.js.map ",
        )


def create_app(title, version='1.0', openapi_tags=None, favicon_url='icon.png') -> FastAPI:
    if openapi_tags is None:
        openapi_tags = [
            {"name": "private",
             "description": "私有接口"},
            {"name": "public",
             "description": "公有接口"},
        ]
    app = FastAPI(openapi_tags=openapi_tags,
                  title=title,
                  version=version,
                  description="",
                  docs_url=None,
                  redoc_url=None,
                  default_response_class=ORJSONResponse,
                  )
    set_app_doc_static_file(app, favicon_url=favicon_url)
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    return app


def add_router(app: FastAPI, router: APIRouter):
    app.include_router(router)



from recordclass import recordclass, RecordClass, make_dataclass, dataobject
# Point = recordclass('Point', 'x y')
# p = Point(1,2)
# print(p)
# Point(1, 2)

if __name__ == "__main__":
    from sys import getsizeof as sizeof
    from collections import namedtuple
    import sys
    sysversion = sys.version
    print(type(sysversion), sysversion.split('\n'))
    # print(sys.platform)

    class Point(RecordClass):
        x: int
        y: int
    p = Point(1, 2,)
    print(p.x, p.y)
    print(p)
    p.x = 3
    print(p)
    print(sys.getsizeof(p))



#
#   Hello World client in Python
#   Connects REQ socket to tcp://localhost:5555
#   Sends "Hello" to server, expects "World" back
#
import zmq

context = zmq.Context()

#  Socket to talk to server
print("Connecting to hello world server…")
socket = context.socket(zmq.REQ)
socket.connect("tcp://localhost:5555")

#  Do 10 requests, waiting each time for a response
for request in range(10):
    print("Sending request %s …" % request)
    socket.send(b"Hello")

    #  Get the reply.
    message = socket.recv()
    print("Received reply %s [ %s ]" % (request, message))



import os
from sparrow.path import rel_to_abs
from glob import glob
import shutil

proto_dir = rel_to_abs('.', return_str=True)
proto_list = [proto for proto in glob(os.path.join(proto_dir, '*.proto'))]


def build_python():
    python_out = os.path.join(proto_dir, 'python')
    for proto_path in proto_list:
        os.system(f"protoc --proto_path={proto_dir} --python_out={python_out} {proto_path}")
        os.system(f"python -m grpc_tools.protoc --proto_path={proto_dir} \
        --python_out={python_out} \
        --grpc_python_out={python_out} \
        {proto_path} ")


def build_web():
    # https://github.com/grpc/grpc-web
    javascript_out = os.path.join(proto_dir, 'js')
    web_proto_path_list = glob(rel_to_abs('../../web/apps/*/src/proto', return_str=True))
    PROTOC_GEN_TS_PATH = os.path.join(rel_to_abs('../../web/apps/chatroom', return_str=True),
                                      "node_modules/.bin/protoc-gen-ts")

    def gen_web_1(proto_path):
        # 可
        os.system(f"protoc --proto_path={proto_dir} \
        --js_out=import_style=commonjs,binary:{javascript_out} \
        --ts_out={javascript_out} \
        --plugin=protoc-gen-ts={PROTOC_GEN_TS_PATH} \
        {proto_path}")

    def gen_web__js(proto_path):
        # 可
        os.system(f"protoc --proto_path={proto_dir} \
        --js_out=import_style=commonjs,binary:{javascript_out} \
        --grpc-web_out=import_style=commonjs,mode=grpcwebtext:{javascript_out} \
        {proto_path}")

    def gen_web__ts(proto_path):
        # 可
        command = f"protoc --proto_path={proto_dir} \
        --js_out=import_style=commonjs:{javascript_out} \
        --grpc-web_out=import_style=typescript,mode=grpcweb:{javascript_out} \
        {proto_path}"
        os.system(command)

    for proto_path in proto_list:
        # gen_web_1(proto_path)
        gen_web__ts(proto_path)

        [shutil.copy(proto_path, web_proto_path) for web_proto_path in web_proto_path_list]

    # copy to web proto dir
    web_proto_path = rel_to_abs("../../web/apps/chatroom/src/proto/module", return_str=True)
    if os.path.exists(web_proto_path):
        shutil.rmtree(web_proto_path)
    shutil.copytree(f"{javascript_out}", web_proto_path)


def copy_proto_to_web():
    web_proto_path_list = glob(rel_to_abs('../../web/apps/*/src/proto', return_str=True))
    for proto_path in proto_list:
        [shutil.copy(proto_path, web_proto_path) for web_proto_path in web_proto_path_list]
        # for web_proto_path in web_proto_path_list:
        #     shutil.copy(proto_path, web_proto_path)


build_python()
# build_web()
copy_proto_to_web()



"""
A rudimentary URL downloader (like wget or curl) to demonstrate Rich progress bars.
"""

import os.path
import sys
from concurrent.futures import ThreadPoolExecutor
import signal
from functools import partial
from threading import Event
from typing import Iterable
from urllib.request import urlopen

from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskID,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

progress = Progress(
    TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
    BarColumn(bar_width=None),
    "[progress.percentage]{task.percentage:>3.1f}%",
    "•",
    DownloadColumn(),
    "•",
    TransferSpeedColumn(),
    "•",
    TimeRemainingColumn(),
)


done_event = Event()


def handle_sigint(signum, frame):
    done_event.set()


signal.signal(signal.SIGINT, handle_sigint)


def copy_url(task_id: TaskID, url: str, path: str) -> None:
    """Copy data from a url to a local file."""
    progress.console.log(f"Requesting {url}")
    response = urlopen(url)
    # This will break if the response doesn't contain content length
    progress.update(task_id, total=int(response.info()["Content-length"]))
    with open(path, "wb") as dest_file:
        progress.start_task(task_id)
        for data in iter(partial(response.read, 32768), b""):
            dest_file.write(data)
            progress.update(task_id, advance=len(data))
            if done_event.is_set():
                return
    progress.console.log(f"Downloaded {path}")


def download(urls: Iterable[str], dest_dir: str):
    """Download multiple files to the given directory."""

    with progress:
        with ThreadPoolExecutor(max_workers=4) as pool:
            for url in urls:
                filename = url.split("/")[-1]
                dest_path = os.path.join(dest_dir, filename)
                task_id = progress.add_task("download", filename=filename, start=False)
                pool.submit(copy_url, task_id, url, dest_path)


if __name__ == "__main__":
    # Try with https://releases.ubuntu.com/20.04/ubuntu-20.04.3-desktop-amd64.iso
    url = "Try with https://releases.ubuntu.com/20.04/ubuntu-20.04.3-desktop-amd64.iso"
    download(url, "./")


import socket


def get_inner_ip():
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        s.connect(('8.8.8.8', 80))
        return s.getsockname()[0]


def get_outer_ip():
    import requests
    # headers = {
    #     'User-Agent':
    #     "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.102 Safari/537.36"
    # }
    # return requests.get('http://ip.42.pl/raw', headers=headers).text.strip()
    return requests.get('http://ifconfig.me/ip', timeout=1).text.strip()


def get_ip(env="inner"):
    """Get in-net / out-ner ip address.
    env: "inner" or "outer"
    """
    if env == "inner":
        return get_inner_ip()
    elif env == "outer":
        return get_outer_ip()
    else:
        raise ValueError("`env` invalid!")


if __name__ == "__main__":
    print(get_inner_ip())
    print(get_outer_ip())



import os
import sys
import shutil
from glob import glob
from sparrow.path import rel_to_abs
import pickle
from typing import Union, List, Dict
from .core import broadcast


@broadcast
def rm(PATH):
    """Enhanced rm, support for regular expressions"""

    def _rm(path):
        """remove path"""
        if os.path.exists(path):
            if os.path.isfile(path):
                os.remove(path)
            elif os.path.isdir(path):
                shutil.rmtree(path)
            else:
                print(f"{path} is illegal.")

    path_list = glob(PATH)
    for path in path_list:
        _rm(path)


def path(string: str) -> str:
    """Adaptive to different platforms"""
    platform = sys.platform.lower()
    if platform in ("linux", "darwin"):
        return string.replace("\\", "/")
    elif platform in ("win", "win32"):
        return string.replace("/", "\\")
    else:
        return string


def save(filename, file):
    with open(filename, "wb") as fw:
        pickle.dump(file, fw)


def load(filename):
    with open(filename, "rb") as fi:
        file = pickle.load(fi)
    return file


def json_load(filepath: str, rel=False):
    import orjson
    abs_path = rel_to_abs(filepath, parents=1) if rel else filepath
    with open(abs_path, 'rb') as f:
        return orjson.loads(f.read())


def json_dump(data: Union[List, Dict], filepath: str, rel=False, indent_2=False):
    import orjson
    orjson_option = 0
    if indent_2:
        orjson_option = orjson.OPT_INDENT_2
    abs_path = rel_to_abs(filepath, parents=1) if rel else filepath
    with open(abs_path, 'wb') as f:
        f.write(orjson.dumps(data, option=orjson_option))

def yaml_dump(filepath, data, rel_path=False):
    abs_path = rel_to_abs(filepath, parents=1) if rel_path else filepath
    from yaml import dump

    try:
        from yaml import CDumper as Dumper
    except ImportError:
        from yaml import Dumper
    with open(abs_path, "w", encoding="utf-8") as fw:
        fw.write(dump(data, Dumper=Dumper, allow_unicode=True, indent=4))


def yaml_load(filepath, rel_path=False):
    abs_path = rel_to_abs(filepath, parents=1) if rel_path else filepath
    from yaml import load

    try:
        from yaml import CLoader as Loader
    except ImportError:
        from yaml import Loader
    with open(abs_path, "r", encoding="utf-8") as stream:
        #     stream = stream.read()
        content = load(stream, Loader=Loader)
    return content



import torch
from torchvision import models
from torchvision.models.efficientnet import EfficientNet_B7_Weights
from torchvision import transforms


def get_model():
    torch_model = models.efficientnet_b7(weights=EfficientNet_B7_Weights.DEFAULT)
    torch_model = torch.nn.Sequential(*(list(torch_model.children())[:-1]))
    torch_model.to('cuda' if torch.cuda.is_available() else 'cpu')
    torch_model.eval()
    return torch_model


preprocess = transforms.Compose([
    transforms.Resize(size=224, interpolation=transforms.InterpolationMode.BICUBIC),
    transforms.ToTensor(),
    transforms.Normalize(
        mean=[0.485, 0.456, 0.406],
        std=[0.229, 0.224, 0.225]
    )
])



import numpy as np
from typing import Callable, Iterable, Sequence
from numpy.typing import ArrayLike


def line_intersection(
        line1: Sequence[Sequence[float]],
        line2: Sequence[Sequence[float]]
) -> np.ndarray:
    """
    return intersection point of two lines,
    each defined with a pair of vectors determining
    the end points
    """
    x_diff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    y_diff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(x_diff, y_diff)
    if div == 0:
        raise Exception("Lines do not intersect")
    d = (det(*line1), det(*line2))
    x = det(d, x_diff) / div
    y = det(d, y_diff) / div
    return np.array([x, y, 0])


def find_intersection(
        p0: ArrayLike,
        v0: ArrayLike,
        p1: ArrayLike,
        v1: ArrayLike,
        threshold: float = 1e-5
) -> np.ndarray:
    """
    Return the intersection of a line passing through p0 in direction v0
    with one passing through p1 in direction v1.  (Or array of intersections
    from arrays of such points/directions).
    For 3d values, it returns the point on the ray p0 + v0 * t closest to the
    ray p1 + v1 * t
    """
    p0 = np.array(p0, ndmin=2)
    v0 = np.array(v0, ndmin=2)
    p1 = np.array(p1, ndmin=2)
    v1 = np.array(v1, ndmin=2)
    m, n = np.shape(p0)
    assert (n in [2, 3])

    numer = np.cross(v1, p1 - p0)
    denom = np.cross(v1, v0)
    if n == 3:
        d = len(np.shape(numer))
        new_numer = np.multiply(numer, numer).sum(d - 1)
        new_denom = np.multiply(denom, numer).sum(d - 1)
        numer, denom = new_numer, new_denom

    denom[abs(denom) < threshold] = np.inf  # So that ratio goes to 0 there
    ratio = numer / denom
    ratio = np.repeat(ratio, n).reshape((m, n))
    return p0 + ratio * v0


def get_closest_point_on_line(
        a: np.ndarray,
        b: np.ndarray,
        p: np.ndarray
) -> np.ndarray:
    """
    It returns point x such that
    x is on line ab and xp is perpendicular to ab.
    If x lies beyond ab line, then it returns nearest edge(a or b).
    """
    # x = b + t*(a-b) = t*a + (1-t)*b
    t = np.dot(p - b, a - b) / np.dot(a - b, a - b)
    if t < 0:
        t = 0
    if t > 1:
        t = 1
    return ((t * a) + ((1 - t) * b))


def cross2d(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    if len(a.shape) == 2:
        return a[:, 0] * b[:, 1] - a[:, 1] * b[:, 0]
    else:
        return a[0] * b[1] - b[0] * a[1]


def tri_area(
        a: Sequence[float],
        b: Sequence[float],
        c: Sequence[float]
) -> float:
    return 0.5 * abs(
        a[0] * (b[1] - c[1]) +
        b[0] * (c[1] - a[1]) +
        c[0] * (a[1] - b[1])
    )


def is_inside_triangle(
        p: np.ndarray,
        a: np.ndarray,
        b: np.ndarray,
        c: np.ndarray
) -> bool:
    """
    Test if point p is inside triangle abc
    """
    crosses = np.array([
        cross2d(p - a, b - p),
        cross2d(p - b, c - p),
        cross2d(p - c, a - p),
    ])
    return np.all(crosses > 0) or np.all(crosses < 0)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sparray.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rsparray.proto\x12\x07sparray\x1a\x1cgoogle/protobuf/struct.proto\"A\n\x11\x44\x65nseNdArrayProto\x12\x0e\n\x06\x62uffer\x18\x01 \x01(\x0c\x12\r\n\x05shape\x18\x02 \x03(\r\x12\r\n\x05\x64type\x18\x03 \x01(\t\"C\n\x08\x44ocProto\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0e\n\x06\x62uffer\x18\x02 \x01(\x0c\x12\x0c\n\x04text\x18\x03 \x01(\t\x12\r\n\x05\x64type\x18\x04 \x01(\t\"0\n\rDocArrayProto\x12\x1f\n\x04\x64ocs\x18\x01 \x03(\x0b\x32\x11.sparray.DocProto\"f\n\tChatProto\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03msg\x18\x02 \x01(\t\x12\x0e\n\x06\x62uffer\x18\x03 \x01(\x0c\x12\r\n\x05\x64type\x18\x04 \x01(\t\x12\x1f\n\x07imgInfo\x18\x05 \x01(\x0b\x32\x0e.sparray.Image\"&\n\x05Image\x12\r\n\x05width\x18\x01 \x01(\x02\x12\x0e\n\x06height\x18\x02 \x01(\x02\x62\x06proto3')



_DENSENDARRAYPROTO = DESCRIPTOR.message_types_by_name['DenseNdArrayProto']
_DOCPROTO = DESCRIPTOR.message_types_by_name['DocProto']
_DOCARRAYPROTO = DESCRIPTOR.message_types_by_name['DocArrayProto']
_CHATPROTO = DESCRIPTOR.message_types_by_name['ChatProto']
_IMAGE = DESCRIPTOR.message_types_by_name['Image']
DenseNdArrayProto = _reflection.GeneratedProtocolMessageType('DenseNdArrayProto', (_message.Message,), {
  'DESCRIPTOR' : _DENSENDARRAYPROTO,
  '__module__' : 'sparray_pb2'
  # @@protoc_insertion_point(class_scope:sparray.DenseNdArrayProto)
  })
_sym_db.RegisterMessage(DenseNdArrayProto)

DocProto = _reflection.GeneratedProtocolMessageType('DocProto', (_message.Message,), {
  'DESCRIPTOR' : _DOCPROTO,
  '__module__' : 'sparray_pb2'
  # @@protoc_insertion_point(class_scope:sparray.DocProto)
  })
_sym_db.RegisterMessage(DocProto)

DocArrayProto = _reflection.GeneratedProtocolMessageType('DocArrayProto', (_message.Message,), {
  'DESCRIPTOR' : _DOCARRAYPROTO,
  '__module__' : 'sparray_pb2'
  # @@protoc_insertion_point(class_scope:sparray.DocArrayProto)
  })
_sym_db.RegisterMessage(DocArrayProto)

ChatProto = _reflection.GeneratedProtocolMessageType('ChatProto', (_message.Message,), {
  'DESCRIPTOR' : _CHATPROTO,
  '__module__' : 'sparray_pb2'
  # @@protoc_insertion_point(class_scope:sparray.ChatProto)
  })
_sym_db.RegisterMessage(ChatProto)

Image = _reflection.GeneratedProtocolMessageType('Image', (_message.Message,), {
  'DESCRIPTOR' : _IMAGE,
  '__module__' : 'sparray_pb2'
  # @@protoc_insertion_point(class_scope:sparray.Image)
  })
_sym_db.RegisterMessage(Image)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _DENSENDARRAYPROTO._serialized_start=56
  _DENSENDARRAYPROTO._serialized_end=121
  _DOCPROTO._serialized_start=123
  _DOCPROTO._serialized_end=190
  _DOCARRAYPROTO._serialized_start=192
  _DOCARRAYPROTO._serialized_end=240
  _CHATPROTO._serialized_start=242
  _CHATPROTO._serialized_end=344
  _IMAGE._serialized_start=346
  _IMAGE._serialized_end=384
# @@protoc_insertion_point(module_scope)



from __future__ import annotations
import os
from typing import Union
from pathlib import Path
import chevron
from rich import print
from rich.panel import Panel
from rich.text import Text
from rich.table import Table
import shutil
from ..path import rel_to_abs, rel_path_join


class expand_template:
    """替换单个文件模板

    Example
    -------
    >>> expand_template(
    ...     name='config_gen.py',
    ...     template='./config.template.ts',
    ...     out="../web/apps/chatroom/src/",
    ...     substitutions={
    ...         "xxxkey": "xxxvalue",
    ...         },
    ...     )
    """

    def __init__(self,
                 name: Union[str, Path],
                 out: Union[str, Path],
                 substitutions: dict,
                 template: Union[str, Path],
                 stdout=False):
        self.name = name if type(name) == "str" else Path(name)
        self.out = out
        self.template = template if type(template) == 'str' else Path(template)
        self.substitutions = substitutions
        self.show = stdout
        self._gen_code()

    @staticmethod
    def write_targe_path(target_path: str, content: str):
        with open(target_path, 'w') as f:
            f.write(content)

    def _gen_code(self):
        with open(rel_to_abs(self.template), 'r') as f:
            try:
                result = chevron.render(f, self.substitutions)
            except Exception as e:
                # 非utf8格式跳过
                print(f"substitute {self.template} failed, skip")
                result = f.read()
            if self.show:
                table = Table(title="", box=None)
                table.add_row(Text("\n".join(result.split('\n')[:10]), tab_size=4))
                table.add_row("...")
                print(Panel(table, title=self.template.name, expand=False, ))
            out_path = rel_path_join(self.out, self.name, return_str=False)
            if not out_path.parent.exists():
                os.makedirs(out_path.parent)
            self.write_targe_path(out_path, result)


class Substituter:
    def __init__(self, stdout=True):
        self._root_dir = None
        self._stdout = stdout

    def render(self, src_dir: str | Path, out_dir: str | Path, substitutions: dict,
               exclude: list | None = None):
        """将`src_dir`中所有文件替换`substitutions`中的key为value
        并输出相同目录树到`out_dir`目录。

        Parameters
        ----------
        src_dir : str or Path
            源目录; 递归遍历指定文件目录下的每个文件目录
        out_dir : str or Path
            输出目录; 递归过程中保持不变
        substitutions : dict
            替换字典; 对文件目录中的每个文件执行替换
        exclude: list or None
            路径中无需替换的文件名/路径

        Examples
        --------
        >>> sutt = Substituter()
        ... sutt.render("./src", "./out", {"package_name": "sparrow"})
        """
        src_dir = Path(src_dir) if type(src_dir) == 'str' else src_dir
        if exclude is None:
            exclude = [".so", ".pyc"]
        if self._root_dir is None:
            self._root_dir = src_dir
        for i in src_dir.iterdir():
            if i.is_dir():
                self.render(i, out_dir, substitutions)
            elif any([i.name.endswith(_) for _ in exclude]):
                print(f"skipping:{str(i)}")
                continue
            else:
                rel_name = i.relative_to(self._root_dir)
                expand_template(
                    name=rel_name,
                    out=out_dir,
                    template=i,
                    substitutions=substitutions,
                    stdout=self._stdout
                )

    @staticmethod
    def render_dir(src_name: str | Path, target_name: str | Path):
        p1 = Path(src_name)
        p2 = Path(target_name)
        if p2.exists():
            p2_file = str(p2)
            p2_file_bk = f"{p2_file}.bk"
            shutil.move(p2_file, p2_file_bk)
            print(f"Warning: Path {p2_file} \nexists! Backup to {p2_file_bk}")
        p1.rename(p2)



from sparrow.template.core import Substituter
from sparrow.path import rel_to_abs
from sparrow.cli import print_directory_tree
from pathlib import Path
from rich import print
import os


def create_project(project_name="my-project", verbose=True):
    src_path = rel_to_abs('src', return_str=False)
    out_path = Path(f"./{project_name}")
    substituter = Substituter(stdout=False)
    substituter.render(src_path,
                       out_path.absolute(),
                       substitutions={"project_name": project_name},
                       exclude=[".so", ".pyc"]
                       )
    substituter.render_dir(os.path.join(out_path, "project_name"),
                           os.path.join(out_path, project_name))
    print(f"project generate success!")
    if verbose:
        print_directory_tree(out_path)


if __name__ == "__main__":
    create_project()



from sparrow.string.color_string import rgb_string, color_const
from git import Repo, RemoteProgress
import os
from sparrow.utils.cursor import Cursor

cursor = Cursor()


class CloneProgress(RemoteProgress):
    MESSAGE = ''

    def update(self, op_code, cur_count, max_count=None, message=""):
        percent = cur_count / (max_count or 100.0) * 100
        self.MESSAGE = message if message != "" else self.MESSAGE
        print(f"\r{int(cur_count)}|{int(max_count)} {self.MESSAGE} PERCENT:{percent:.0f}% {cursor.EraseLine(0)} ",
              end='', flush=True)


def clone(url: str, repo_name=None, branch=None):
    url = url.strip()
    if repo_name is None:
        repo_name = url.split('/')[-1][:-len('.git')]
    if url.startswith('git'):
        git_proxy_dir = url
    else:
        git_proxy_dir = f"https://ghproxy.com/{url.strip()}"
    print(rgb_string(f"Cloning into '{repo_name}' ...", color=color_const.green))
    Repo.clone_from(git_proxy_dir, to_path=repo_name, branch=branch, progress=CloneProgress())


def clone_cmd(url: str):
    os.system(f"git clone https://ghproxy.com/{url.strip()}")


if __name__ == "__main__":
    clone("https://github.com/beidongjiedeguang/beidongjiedeguang.git")
    # clone_cmd("https://github.com/beidongjiedeguang/beidongjiedeguang.git")



from functools import partial, wraps, reduce
import inspect
import numpy as np
import operator as op
import random
import time
from .utils import exists

__all__ = ["topk", "dict_topk", "random_idx", "clamp", "get_num_args", "get_parameters"]


def dict_topk(a: dict, k: int, reverse=False):
    from pandas import DataFrame
    df = DataFrame({'key': a.keys(), 'value': a.values()})
    if not reverse:
        return df.nlargest(k, 'value')
    else:
        return df.nsmallest(k, 'value')


def topk(a, k, axis=-1, largest=True, sort=True):
    """Series top K"""
    a = np.asanyarray(a)
    if axis is None:
        axis_size = a.size
    else:
        axis_size = a.shape[axis]
    assert 1 <= k <= axis_size

    if largest:
        index_array = np.argpartition(a, axis_size - k, axis=axis)
        topk_indices = np.take(index_array, -np.arange(k) - 1, axis=axis)
    else:
        index_array = np.argpartition(a, k - 1, axis=axis)
        topk_indices = np.take(index_array, np.arange(k), axis=axis)
    topk_values = np.take_along_axis(a, topk_indices, axis=axis)
    if sort:
        sorted_indices_in_topk = np.argsort(topk_values, axis=axis)
        if largest:
            sorted_indices_in_topk = np.flip(sorted_indices_in_topk, axis=axis)
        sorted_topk_values = np.take_along_axis(
            topk_values, sorted_indices_in_topk, axis=axis
        )
        sorted_topk_indices = np.take_along_axis(
            topk_indices, sorted_indices_in_topk, axis=axis
        )
        return sorted_topk_values, sorted_topk_indices
    return topk_values, topk_indices


def random_idx(idx_range, exclude_idx=None):
    random.seed(time.time())
    rand_idx = random.randint(*idx_range)
    if rand_idx == exclude_idx:
        return random_idx(idx_range, exclude_idx)
    else:
        return rand_idx


def clamp(x, x_min=None, x_max=None):
    """Clamp a number to same range.
    Examples:
        >>> clamp(-1, 0, 1)
        >>> 0
        >>> clamp([-1, 2, 3], [0, 0, 0], [1, 1, 1])
        >>> [0, 1, 1]
    """
    assert exists(x_min) or exists(x_max)
    if exists(x_min):
        x = np.maximum(x, x_min)
    if exists(x_max):
        x = np.minimum(x, x_max)
    return x


CHOOSE_CACHE = {}


def choose_using_cache(n, r):
    if n not in CHOOSE_CACHE:
        CHOOSE_CACHE[n] = {}
    if r not in CHOOSE_CACHE[n]:
        CHOOSE_CACHE[n][r] = choose(n, r, use_cache=False)
    return CHOOSE_CACHE[n][r]


def choose(n, r, use_cache=True):
    if use_cache:
        return choose_using_cache(n, r)
    if n < r:
        return 0
    if r == 0:
        return 1
    denom = reduce(op.mul, range(1, r + 1), 1)
    numer = reduce(op.mul, range(n, n - r, -1), 1)
    return numer // denom


def get_num_args(function):
    return len(get_parameters(function))


def get_parameters(function):
    return inspect.signature(function).parameters



from .io import yaml_load, yaml_dump, save, load, rm, json_load, json_dump
from .decorators import benchmark
from .string.color_string import rgb_string
from .functions.core import clamp
from .path import rel_to_abs
from .functions.core import topk, dict_topk
from .progress_bar import probar

_version_config = yaml_load("version-config.yaml", rel_path=True)
__version__ = _version_config["version"]
# print(f"{rgb_string(_version_config['name'], color='#34A853')} version: {rgb_string(__version__, color='#C9E8FF')}")



from ..string.color_string import rgb_string, color_const
from ..log.core import BaseLogger, SimpleLogger
from functools import wraps, update_wrapper
import logging
import inspect
import time
import types
from typing import Union


class TryDecorator:
    def __init__(self, error_return=None, logger: Union[SimpleLogger, None] = None):
        self.error_return = error_return
        self.logger = logger

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                result = func(*args, **kwargs)
            except Exception as e:
                if self.logger:
                    self.logger.warning(f"{func.__name__} raised an exception: {e}")
                else:
                    print(f"{func.__name__} raised an exception: {e}")
                result = self.error_return
            return result

        return wrapper

    def __get__(self, instance, cls):
        if instance is None:
            return self
        else:
            return types.MethodType(self, instance)


def try_decorator(error_return=None, logger: Union[SimpleLogger, None] = None):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                value = func(*args, **kwargs)
            except Exception as e:
                if logger:
                    logger.warning(f"{func.__name__} error: {e}")
                else:
                    print(f"{func.__name__} error: {e}")
                return error_return
            return value

        return wrapper

    return decorator


def benchmark(times=10, logger: Union[SimpleLogger, None] = None, level=logging.INFO, ):
    # if func is None:
    #     return partial(time_it, times=times)
    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start = time.time()
            value = None
            for i in range(times):
                value = func(*args, **kwargs)
            end = time.time()
            average_cost_time = (end - start) / times
            time_str = f"{average_cost_time:.3f}"

            if logger:
                logger.log(
                    level,
                    f"Run {rgb_string(str(times), color_const.GREEN)} times, "
                    f"the average time is {rgb_string(time_str, color_const.GREEN)} seconds."
                )
            else:
                print(
                    f"Run {rgb_string(str(times), color_const.GREEN)} times, "
                    f"the average time is {rgb_string(time_str, color_const.GREEN)} seconds."
                )
            return value

        return wrapper

    return decorate


def measure_time(logger: Union[SimpleLogger, None] = None, level=logging.INFO):
    def decorate(func):
        """Log the runtime of the decorated function."""

        @wraps(func)
        def wrapper(*args, **kwargs):
            start = time.time()
            value = func(*args, **kwargs)
            end = time.time()
            cost_time = end - start
            time_str = f"{cost_time:.3f}"
            if logger:
                logger.log(level,
                           f"Finished {rgb_string(func.__name__, color_const.RED)} in {rgb_string(time_str, color_const.GREEN)} secs."
                           )
            else:
                print(f"Finished {rgb_string(func.__name__, color_const.RED)} in {rgb_string(time_str, color_const.GREEN)} secs.")
            return value

        return wrapper

    return decorate


def repeat(n=2):
    """repeat decorated function `n` times."""

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = None
            for i in range(n):
                result = func(*args, **kwargs)
            return result

        return wrapper

    return decorator


def optional_debug(func):
    if "debug" in inspect.signature(func).parameters:
        raise TypeError("debug argument already defined")

    debug_default = True

    @wraps(func)
    def wrapper(*args, debug=debug_default, **kwargs):
        if debug:
            args_repr = [repr(a) for a in args]
            kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
            signature = ", ".join(args_repr + kwargs_repr)
            print(f"Calling '{func.__name__}({signature})'")
        value = func(*args, **kwargs)
        if debug:
            print(f"{func.__name__!r} returned {value!r}")
        return value

    sig = inspect.signature(func)
    parms = list(sig.parameters.values())
    parms.append(
        inspect.Parameter(
            "debug", inspect.Parameter.KEYWORD_ONLY, default=debug_default
        )
    )
    wrapper.__signature__ = sig.replace(parameters=parms)
    return wrapper


def count_calls(func):
    """Count the number of calls made to the decorated function."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        wrapper.num_calls += 1
        print(f"Call {wrapper.num_calls} of {func.__name__!r}")
        return func(*args, **kwargs)

    wrapper.num_calls = 0
    return wrapper


class CountCalls:
    """Count the number of calls made to the decorated function."""

    def __init__(self, func):
        update_wrapper(self, func)
        self.func = func
        self.num_calls = 0

    def __call__(self, *args, **kwargs):
        self.num_calls += 1
        print(f"Call {self.num_calls} of {self.func.__name__!r}")
        return self.func(*args, **kwargs)


class CallReminder:
    def __init__(self, func):
        self._func = func
        self._num_calls = 0

    def __call__(self, *args, **kwargs):
        self._num_calls += 1
        return self._func(*args, **kwargs)

    @property
    def count_calls(self):
        return self._num_calls


def logged(level, name=None, message=None):
    """
    Add logging to a function. level is the logging
    level, name is the logger name, and message is the
    log message. If name and message aren't specified,
    they default to the function's module and name.
    """

    def decorate(func):
        logname = name if name else func.__module__
        log = logging.getLogger(logname)
        logmsg = message if message else func.__name__

        @wraps(func)
        def wrapper(*args, **kwargs):
            log.log(level, logmsg)
            return func(*args, **kwargs)

        return wrapper

    return decorate


def singleton(cls):
    """
    Usage
    -----
        @singleton
        class Cls:
            pass
    """
    _instance = {}

    def inner():
        if cls not in _instance:
            _instance[cls] = cls()
        return _instance[cls]

    return inner


class Singleton:
    """
    Usage
    -----
        @Singleton
        class Cls:
            pass
    """

    def __init__(self, cls):
        self._cls = cls
        self._instance = {}

    def __call__(self):
        if self._cls not in self._instance:
            self._instance[self._cls] = self._cls()
        return self._instance[self._cls]


class MetaSingleton(type):
    """
    Usage
    -----
        class Cls(metaclass=MetaSingleton):
            pass
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(MetaSingleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



from rich.live import Live
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn, BarColumn, TextColumn
from rich.table import Table


class probar:
    def __init__(self, iterable, total_steps=None, title=None):
        """
        """
        self.iterable = iterable
        if total_steps is None:
            try:
                total_steps = len(iterable)
                if not total_steps:
                    total_steps = None
            except:
                total_steps = None
        self.total_steps = total_steps
        self.title = title

    def __iter__(self):
        with Progress() as progress:
            task = progress.add_task(self.title, total=self.total_steps)
            for idx, item in enumerate(self.iterable):
                yield item
                progress.update(task, advance=1)


class MultiProbar:
    def __init__(self):
        self._progress = Progress()
        self._overall_progress = Progress()
        self.live_progress = None

    def add_task(self, description, total):
        self._progress.add_task(description=description,
                                total=total)

    def start(self):
        # self._progress.start()
        total = sum(task.total for task in self._progress.tasks)
        overall_task = self._overall_progress.add_task("All Jobs", total=int(total))

        progress_table = Table.grid()
        progress_table.add_row(
            Panel.fit(
                self._overall_progress, title="Overall Progress", border_style="green", padding=(2, 2)
            ),
            Panel.fit(self._progress, title="[b]Jobs", border_style="red", padding=(1, 2)),
        )

        self.progress_table = progress_table
        self.live_progress = Live(progress_table, refresh_per_second=10)
        self.live_progress.start()

        # with Live(progress_table, refresh_per_second=10):
        #     while not self._overall_progress.finished:
        #         # progress
        #         time.sleep(0.1)
        #         for job in self._progress.tasks:
        #             if not job.finished:
        #                 self._progress.advance(job.id)
        #
        #         completed = sum(task.completed for task in self._progress.tasks)
        #         self._overall_progress.update(overall_task, completed=completed)

    def update(self, advance=1):
        if not self._overall_progress.finished:
            time.sleep(0.1)
            for job in self._progress.tasks:
                if not job.finished:
                    self._progress.advance(job.id, advance=advance)

            completed = sum(task.completed for task in self._progress.tasks)
            self._overall_progress.update(self.progress_table, completed=completed)

    def stop(self):
        self._progress.stop()


if __name__ == "__main__":
    import time
    import threading

    # for i in probar(range(100), title="emmm"):
    #     time.sleep(0.1)

    def test1():
        def show(progress, task, total: int):
            for i in range(total):
                progress.update(task, advance=1)
                time.sleep(.005)

        progress = Progress()
        progress.start()
        task1 = progress.add_task(description=str(1000), total=1000)
        task2 = progress.add_task(description=str(2000), total=2000)
        t1 = threading.Thread(target=show, name='t1', args=(progress, task1, 1000))
        t2 = threading.Thread(target=show, name='t2', args=(progress, task2, 2000))
        t1.start()
        t2.start()
        [i.join() for i in [t1, t2]]
        progress.stop()
        print("test1 complete")


    def test2():
        with Progress() as progress:
            task1 = progress.add_task("[red]Downloading...", total=1000)
            task2 = progress.add_task("[green]Processing...", total=100)
            task3 = progress.add_task("[cyan]Cooking...", total=1000)

            while not progress.finished:
                progress.update(task1, advance=0.5)
                progress.update(task2, advance=0.3)
                progress.update(task3, advance=0.9)
                time.sleep(0.02)



    def test3():
        from time import sleep

        job_progress = Progress(
            "{task.description}",
            SpinnerColumn(),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        )
        job1 = job_progress.add_task("[green]Cooking")
        job2 = job_progress.add_task("[magenta]Baking", total=200)
        job3 = job_progress.add_task("[cyan]Mixing", total=400)

        total = sum(task.total for task in job_progress.tasks)
        overall_progress = Progress()
        overall_task = overall_progress.add_task("All Jobs", total=int(total))

        progress_table = Table.grid()
        progress_table.add_row(
            Panel.fit(
                overall_progress, title="Overall Progress", border_style="green", padding=(2, 2)
            ),
            Panel.fit(job_progress, title="[b]Jobs", border_style="red", padding=(1, 2)),
        )

        with Live(progress_table, refresh_per_second=10):
            while not overall_progress.finished:
                sleep(0.1)
                for job in job_progress.tasks:
                    if not job.finished:
                        job_progress.advance(job.id)

                completed = sum(task.completed for task in job_progress.tasks)
                overall_progress.update(overall_task, completed=completed)




    test1()
    # test2()
    # test3()
    # test4()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: coordinate.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x63oordinate.proto\x12\x0fprotocoordinate\"\xee\x01\n\nUpdateData\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x31\n\x06\x63oords\x18\x02 \x03(\x0b\x32!.protocoordinate.UpdateData.Coord\x12;\n\x0bquaternions\x18\x03 \x03(\x0b\x32&.protocoordinate.UpdateData.Quaternion\x1a(\n\x05\x43oord\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\x1a\x38\n\nQuaternion\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\x12\t\n\x01w\x18\x04 \x01(\x02\x62\x06proto3')



_UPDATEDATA = DESCRIPTOR.message_types_by_name['UpdateData']
_UPDATEDATA_COORD = _UPDATEDATA.nested_types_by_name['Coord']
_UPDATEDATA_QUATERNION = _UPDATEDATA.nested_types_by_name['Quaternion']
UpdateData = _reflection.GeneratedProtocolMessageType('UpdateData', (_message.Message,), {

  'Coord' : _reflection.GeneratedProtocolMessageType('Coord', (_message.Message,), {
    'DESCRIPTOR' : _UPDATEDATA_COORD,
    '__module__' : 'coordinate_pb2'
    # @@protoc_insertion_point(class_scope:protocoordinate.UpdateData.Coord)
    })
  ,

  'Quaternion' : _reflection.GeneratedProtocolMessageType('Quaternion', (_message.Message,), {
    'DESCRIPTOR' : _UPDATEDATA_QUATERNION,
    '__module__' : 'coordinate_pb2'
    # @@protoc_insertion_point(class_scope:protocoordinate.UpdateData.Quaternion)
    })
  ,
  'DESCRIPTOR' : _UPDATEDATA,
  '__module__' : 'coordinate_pb2'
  # @@protoc_insertion_point(class_scope:protocoordinate.UpdateData)
  })
_sym_db.RegisterMessage(UpdateData)
_sym_db.RegisterMessage(UpdateData.Coord)
_sym_db.RegisterMessage(UpdateData.Quaternion)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _UPDATEDATA._serialized_start=38
  _UPDATEDATA._serialized_end=276
  _UPDATEDATA_COORD._serialized_start=178
  _UPDATEDATA_COORD._serialized_end=218
  _UPDATEDATA_QUATERNION._serialized_start=220
  _UPDATEDATA_QUATERNION._serialized_end=276
# @@protoc_insertion_point(module_scope)



import sparrow.multiprocess.server as multi_server
from sparrow.multiprocess.config import Config
from copy import copy


class Client:
    """
    Example:
    First start service:
    ~$ sparrow start-server
    >>> from sparrow.multiprocess.client import Client
    >>> client = Client()
    >>> client.update_dict({'a': 1, 'b': 2})
    >>> print(client.get_dict_data())
    """

    def __init__(self, port=None):
        server_config = Config()
        if port:
            server_config.port = port
        multi_server.register_client_methos()
        self._manager = multi_server.DataManager(
            address=(server_config.host, server_config.port),
            authkey=server_config.authkey)
        self._manager.connect()
        self._dict = self._manager.get_dict()
        self._list = self._manager.get_list()
        self._queue = self._manager.get_queue()
        self._deque = self._manager.get_deque()

    def update_dict(self, data: dict):
        self._dict.update(data)

    def append_to_list(self, data):
        self._list.append(data)

    def put_queue(self, data):
        self._queue.put(data)

    def append_deque(self, data):
        self._deque.append(data)

    def get_deque(self):
        return self._deque

    def get_dict_data(self):
        """Return a copy of original data."""
        return self._dict

    def get_list_data(self):
        return self._list

    def get_queue(self, block=True):
        try:
            return self._queue.get(block=block)
        except Exception as e:
            return None

    def get_raw_data(self):
        return self._dict


if __name__ == "__main__":
    import numpy as np

    client = Client()
    client.put_queue(['asdkf', 'asdf'])
    client.append_deque(['asdf'])
    client.append_deque(['asdf'])
    client.append_deque(['asdf'])
    client.update_dict({'a': 1, 'b': 100})
    client.append_to_list(np.sin(np.linspace(0, 10, 10)))
    print(client.get_deque())
    while True:
        print(client.get_queue(block=True))
        print(client.get_dict_data())
        print(client.get_list_data())
        print(client.get_deque())
        print("------------------------")



import numpy as np
from PIL.Image import fromarray
from IPython import get_ipython
from IPython.display import display_html


def display_np_arrays():
    def np_to_png(a):
        if 2 <= len(a.shape) <= 3:
            return fromarray(
                np.array(np.clip(a, 0, 1) * 255, dtype="uint8")
            )._repr_png_()
        else:
            return fromarray(np.zeros([1, 1], dtype="uint8"))._repr_png_()

    def np_to_text(obj, p, cycle):
        if len(obj.shape) < 2:
            print(repr(obj))
        if 2 <= len(obj.shape) <= 3:
            pass
        else:
            print("<array of shape {}>".format(obj.shape))

    get_ipython().display_formatter.formatters["image/png"].for_type(
        np.ndarray, np_to_png
    )
    get_ipython().display_formatter.formatters["text/plain"].for_type(
        np.ndarray, np_to_text
    )


_style_inline = """<style>
.sparrow-answer {
    color: transparent;
    padding: 5px 15px;
    background-color: #def;
}
.sparrow-answer:hover { color: blue; } 
</style>
"""


def guess(x):
    display_html(
        _style_inline
        + "<h4>Answer is: <span class='einops-answer'>{x}</span> (hover to see)</h4>".format(
            x=tuple(x)
        ),
        raw=True,
    )



from collections import Counter
from functools import wraps
import pickle
import numpy as np
from deprecated import deprecated


def clamp(x, x_min, x_max):
    """Clamp a number to same range.
    It's equivalent to np.clip()
    Examples:
        >>> clamp(-1, 0, 1)
        >>> 0
        >>> clamp([-1, 2, 3], [0, 0, 0], [1, 1, 1])
        >>> [0, 1, 1]
    """
    return np.maximum(x_min, np.minimum(x_max, x))


def broadcast(func):  # It can be replaced by `np.vectorize`
    """Only for a functions with a single argument
    example:
    @broadcast
    def f(x):
        # A function that can map only a single element
        if x==1 or x==0:
            return x
        else:
            return f(x-1)+f(x-2)

    >> f([2,4,10])
    >> (1, 3, 832040)
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        value_list = []
        for arg in args:
            value_list.append(func(arg, **kwargs))
        return tuple(value_list)

    return wrapper


class _Dict_enhance(dict):
    """Enables the dictionary to be dot operated"""

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.__dict__ = self


def dict_dotable(dic):
    """
    : input: a dictionary
    : output: an enhanced dictionary
    Example:
        enhance_dic = dict_dotable(dic)
    then, you can operate an enhanced dictionary like this:
        enhance_dic.key1.key2. ...
    """
    dic = _Dict_enhance(dic)
    for i in dic:
        if type(dic[i]) == dict:
            dic[i] = dict_dotable(dic[i])
    return dic


class Constant:
    """
    define a constant like C language.
    `object.__setattr__(self, name, value)`
    this built-in function will called when assigning values to properties of the class

    `object.__dict__` holds all writable attributes in object,
    key as variable name and value as variable value.
    """

    def __setattr__(self, name, value):
        if hasattr(self, name):
            raise ValueError("Constant value can't be changed")
        else:
            self.__dict__[name] = value


def number_digits(number):
    res = number
    digit = 1
    if res >= 1:
        while res > 10:
            digit += 1
            # res, mod = np.divmod(res, 10)
            res //= 10
    else:
        while res < 1:
            digit -= 1
            res *= 10
    return digit


def num_digits(number_like):
    number_str = str(int(str(number_like)))
    return len(number_str)


def sort_count(_list: list):
    """
    返回lis的由大到小排序好的元素列表
    Example:
    l = np.array([2,2,2,2,5,5,3,9,9])
    sort_count(l) : [(2, 4), (5, 2), (9, 2), (3, 1)]
    # return [2, 5, 9,3], [4, 2, 2, 1]
    """
    a = Counter(_list)
    b = sorted(a.items(), key=lambda item: item[1], reverse=True)
    # idx, counts = [b[i][0] for i in range(len(b))], [b[i][1] for i in range(len(b))]
    return b


def reduce_list_element(array, *elems):
    """
    example:
    a = [ 5, 6, 6, 7, 8, 9, 9]
    reduce_list_element(a, 6, 9)
    print(a)
    >> [ 5, 7, 8]
    """
    length = len(array)
    for idx in range(length):
        index = length - idx - 1
        for elem in elems:
            if array[index] == elem:
                array.pop(index)



import abc
import sys

CSI = '\033['
OSC = '\033]'
OFF = CSI + '0m'


class BaseCursor(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def FORWARD(self, *args, **kwargs):
        '''Forward move'''
        pass

    @abc.abstractmethod
    def BACK(self, *args, **kwargs):
        '''Backward move'''
        pass

    @abc.abstractmethod
    def PreLINE(self, *args, **kwargs):
        """Moves cursor to beginning of the line n (default 1) lines up.
        No any character will be cleared.
        """
        pass

    @abc.abstractmethod
    def EraseLine(self, *args, **kwargs):
        """Erases part of the line.
        If n is 0, clear from cursor to the end of the line.
        If n is 1, clear from cursor to beginning of the line.
        If n is 2, clear entire line. Cursor position does not change.
        """
        pass

    # @abc.abstractmethod
    # def EraseUpLine(self, *args, **kwargs):
    #     """Erase n lines and cursor move up n lines"""
    #     pass

    @abc.abstractmethod
    def savePOS_x(self, *args, **kwargs):
        """Save current position"""
        pass

    @abc.abstractmethod
    def restorePOS_x(self, *args, **kwargs):
        """Restore saved position"""
        pass

    @abc.abstractmethod
    def getPOS(self):
        pass


def send(code):
    sys.stdout.write(code)
    sys.stdout.flush()


def csi(n, code):
    return CSI + str(n) + code


class Cursor(BaseCursor):
    '''See: http://en.wikipedia.org/wiki/ANSI_escape_code'''

    def UP(self, n=1):
        """Moves the cursor n (default 1) cells in the given direction."""
        return csi(n, 'A')

    def DOWN(self, n=1):
        """Moves the cursor n (default 1) cells in the given direction."""
        return csi(n, 'B')

    def FORWARD(self, n=1):
        """Moves the cursor n (default 1) cells in the given direction."""
        return csi(n, 'C')

    def BACK(self, n=1):
        """Moves the cursor n (default 1) cells in the given direction."""
        return csi(n, 'D')

    def NextLINE(self, n=1):
        return csi(n, 'E')

    def PreLINE(self, n=1):
        """Moves cursor to beginning of the line n (default 1) lines up.
        No any character are cleared.
        """
        return csi(n, 'F')

    def HorizCol(self, n=1):
        """Moves the cursor to column n"""
        return csi(n, 'G')

    def POS(self, x=1, y=1):
        """Moves the cursor to row n, column m.
        The values are 1-based, and default to 1 (top left corner) if omitted."""
        return CSI + str(y) + ';' + str(x) + 'H'

    def EraseLine(self, n=0):
        """Erases part of the line.
        If n is 0, clear from cursor to the end of the line.
        If n is 1, clear from cursor to beginning of the line.
        If n is 2, clear entire line. Cursor position does not change.
        """
        if n == 0:
            return CSI + 'K'
        else:
            return csi(n, 'K')

    def EraseDisplay(self, n=0):
        """Clears part of the screen.
        If n is 0, clear from cursor to end of screen.
        If n is 1, clear from cursor to beginning of the screen.
        If n is 2, clear entire screen (and moves cursor to upper left on DOS ANSI.SYS).
        If n is 3, clear entire screen and delete all lines saved in the scrollback buffer.
        """
        return csi(n, 'J')

    def ScrollUp(self, n=1):
        """Scroll whole page up by n (default 1) lines.
        New lines are added at the bottom.
        """
        return csi(n, "S")

    def ScrollDown(self, n=0):
        """Scroll whole page down by n (default 1) lines.
        New lines are added at the top. """
        return csi(n, "T")

    def savePOS_x(self):
        return "\033[s"  # "\0337"

    def restorePOS_x(self):
        return "\033[u"  # "\0338"

    def getPOS(self):
        return CSI + "6n"

    def hiddenCursor(self):
        return "\033[?25l"



from functools import reduce
import inspect
import numpy as np
import scipy.stats as st
import operator as op


def gaussian(x, sigma, miu):
    """Standard Gaussian function"""
    a = 1 / np.sqrt(2 * np.pi * sigma**2)
    return a * np.exp(-((x - miu) ** 2) / (2 * sigma**2))


def norm_gaussian(x, sigma, miu):
    """Normalized Gaussion fuction"""
    return np.exp(-((x - miu) ** 2) / (2 * sigma**2))


def gaussian_2d(U, V, sigma, miu):
    a = 1 / np.sqrt(2 * np.pi * sigma**2)
    res_max = a * np.exp(
        -((0 - miu) ** 2 + (0 - miu) ** 2) / (2 * sigma**2)
    )  # For normalization
    res = a * np.exp(-((U - miu) ** 2 + (V - miu) ** 2) / (2 * sigma**2))
    return res / res_max  # Normalized


def cdf(x, sigma, miu):
    return st.norm.cdf(x, loc=miu, scale=sigma)


def gaussion_kernel(kernlen=3, nsig=3):
    x = np.linspace(-nsig, nsig, kernlen + 1)
    kern1d = np.diff(st.norm.cdf(x))
    kern2d = np.outer(kern1d, kern1d)
    return kern2d / kern2d.sum()


if __name__ == "__main__":
    x = np.repeat(np.arange(5).reshape(1, -1), repeats=5, axis=0)
    y = np.repeat(np.arange(5).reshape(-1, 1), repeats=5, axis=1)

    # print(x)
    # print(y)
    print(gaussian_2d(x, y, sigma=3, miu=2))
    print(gaussion_kernel(5, 3))



from functools import partial, wraps, reduce
from contextlib import contextmanager


def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


def cast_tuple(val, length=1):
    return val if isinstance(val, tuple) else ((val,) * length)


@contextmanager
def null_context(*args, **kwargs):
    yield


def pick_and_pop(keys, d):
    values = list(map(lambda key: d.pop(key), keys))
    return dict(zip(keys, values))


def group_dict_by_key(cond, d):
    return_val = [dict(), dict()]
    for key in d.keys():
        match = bool(cond(key))
        ind = int(not match)
        return_val[ind][key] = d[key]
    return (*return_val,)


def string_begins_with(prefix, str):
    return str.startswith(prefix)


def group_by_key_prefix(prefix, d):
    return group_dict_by_key(partial(string_begins_with, prefix), d)


def groupby_prefix_and_trim(prefix, d):
    kwargs_with_prefix, kwargs = group_dict_by_key(partial(string_begins_with, prefix), d)
    kwargs_without_prefix = dict(map(lambda x: (x[0][len(prefix):], x[1]), tuple(kwargs_with_prefix.items())))
    return kwargs_without_prefix, kwargs


def num_to_groups(num, divisor):
    groups = num // divisor
    remainder = num % divisor
    arr = [divisor] * groups
    if remainder > 0:
        arr.append(remainder)
    return arr


def find_first(cond, arr):
    for el in arr:
        if cond(el):
            return el
    return None



import pandas as pd
from typing import List, Union


def groupby_choice(df: pd.DataFrame, by: Union[str, List], col_name: any, choice='max', inplace=True):
    """
    取分组后的某列最值,组成的新df. 默认inplace.

    Example::
     df = pd.DataFrame({'key' : ['A', 'A', 'B', 'B', 'C', 'C'],
                       'value' : ['v1', 'v2', 'v3', 'v4','v5', 'v6'],
                       'prob' : [1, 5, 50, 2, 5, 5]})
    >>> df
        key value  prob
    0   A    v1     1
    1   A    v2     5
    2   B    v3    50
    3   B    v4     2
    4   C    v5     5
    5   C    v6     5
    >>> groupby_choice(df, 'key', 'prob', 'max')
    >>>
        key value  prob
    1   A    v2     5
    2   B    v3    50
    4   C    v5     5
    """
    if not inplace:
        df = df.copy(deep=True)
    index_list = []
    for idx, item in df.groupby(by)[col_name]:
        if choice == "max":
            index_list.append(item.idxmax())
        elif choice == "min":
            index_list.append(item.idxmin())
        else:
            raise "Invalid `func` parameter."
    return df.iloc[index_list]


def group_df(df, col_name, interval=5, use_max_min_interval=False, closed='neither', dropna=True):
    """
    Parameters
    ----------
        col_name: 根据 `col_name` 进行分组
        interval: 合并采样间隔
        use_max_min_interval: True使用最大最小区间确定等距采样个数； False使用df的样本数目确定采样个数

    """
    if dropna:
        df = df.dropna(axis=0, how='any', inplace=False)
    df = df.sort_values(by=col_name, ascending=True)
    if use_max_min_interval:
        periods = (df[col_name].max() - df[col_name].min()) / interval
    else:
        periods = len(df) // interval

    bins = pd.interval_range(df[col_name].min(), df[col_name].max(),
                             periods=periods,
                             closed=closed)
    pd_cut = pd.cut(df[col_name], bins=bins)
    for idx, i in enumerate(df.groupby(pd_cut)):
        agg_res = i[1].agg('mean')
        if idx == 0:
            df_grouped = agg_res
        else:
            df_grouped = pd.concat([df_grouped, agg_res], axis=1)
    df_grouped = df_grouped.transpose()
    return df_grouped.dropna().reset_index(inplace=False).drop(['index'], axis=1)


def re_ord_df_col(df, col_name, ord_num=0):
    """Re-order df's column name."""
    tmp_list = df.columns.tolist()
    tmp_list.remove(col_name)
    tmp_list.insert(ord_num, col_name)
    df = df[tmp_list]
    return df



from multiprocessing.managers import SyncManager, BaseManager
from typing import TypeVar, List, Dict
from queue import Queue
from collections import deque


# class QueueManager(BaseManager):
#     queue = Queue()
#
#     def get_queue(self) -> Queue:
#         pass
#
#
# DictProxy = TypeVar('DictProxy')


class DataManager(SyncManager):
    data_dict = {}
    data_queue = Queue()
    data_deque = None
    data_list = []

    @classmethod
    def set_data_deque(cls, maxlen):
        cls.data_deque = deque(maxlen=maxlen)

    def get_dict(self) -> Dict:
        ...

    def get_queue(self) -> Queue:
        ...

    def get_deque(self) -> deque:
        ...

    def get_list(self) -> List:
        ...


def register_server_methods():
    DataManager.register('get_dict', callable=lambda: DataManager.data_dict)
    DataManager.register('get_queue', callable=lambda: DataManager.data_queue)
    DataManager.register('get_deque', callable=lambda: DataManager.data_deque)
    DataManager.register('get_list', callable=lambda: DataManager.data_list)


def register_client_methos():
    DataManager.register('get_dict')
    DataManager.register('get_queue')
    DataManager.register('get_deque')
    DataManager.register('get_list')


if __name__ == "__main__":
    from sparrow.multiprocess.config import Config

    config = Config()
    register_server_methods()
    manager = DataManager(address=(config.host, config.port),
                          authkey=config.authkey)
    manager.get_server().serve_forever()



from functools import wraps
import inspect
import os

__all__ = ["get_func_param_dict"]


def get_func_param_dict():
    frame = inspect.currentframe().f_back
    args, varargs, keywords, loc = inspect.getargvalues(frame)
    target_dict = {}
    for key in args:
        target_dict[key] = loc[key]
    return target_dict


# Useless
def findcaller(func):
    @wraps(func)
    def wrapper(*args):
        currentframe = inspect.currentframe()
        f = currentframe.f_back
        file_name = os.path.basename(f.f_code.co_filename)
        func_name = f.f_code.co_name
        line_num = f.f_lineno

        args = list(args)
        args.append(f"{os.path.basename(file_name)}.{func_name}.{line_num}")
        func(*args)

    return wrapper



# from einops import rearrange, reduce
import einops


def repeat(tensor, n, axis=-1):
    shape = tensor.shape
    dims = len(shape)
    if dims == 1:
        if axis == 0:
            res = einops.repeat(tensor, "w -> h w", h=n)
        else:
            res = einops.repeat(tensor, "h -> h w", w=n)
    elif dims == 2:
        if axis == 0:
            res = einops.repeat(tensor, "h w -> c h w", c=n)
        else:
            res = einops.repeat(tensor, "h w -> h w c", c=n)
    elif dims == 3:
        if axis == 0:
            res = einops.repeat(tensor, "h w c -> b h w c", b=n)
        elif axis == 1:
            res = einops.repeat(tensor, "h w c -> h b w c", b=n)
        else:
            res = einops.repeat(tensor, "h w c -> h w c b", b=n)
    else:
        raise ValueError
    return res



import pretty_errors
# from .widgets import timer
from .multiprocess import kill
from .multiprocess import start_server
from .docker import save_docker_images
from .docker import load_docker_images
from .docker.nvidia_stat import docker_gpu_stat
from .utils.compress import pack, unpack
from .proxy import clone
from .net import get_ip
from .template.scaffold.core import create_project
from .ann import milvus


class Cli:
    def __init__(self):
        ...


func_list = [
    # timer,
    save_docker_images,
    load_docker_images,
    docker_gpu_stat,
    pack,
    unpack,
    start_server,
    kill,
    clone,
    get_ip,
    create_project,
    milvus,
]


def fire_commands():
    import fire
    func_dict = {func.__name__: func for func in func_list}
    fire.Fire(func_dict)


def typer_commands():
    import typer
    app = typer.Typer()
    [app.command()(i) for i in func_list]
    app()


def main():
    use_fire = 1
    if use_fire:
        fire_commands()
    else:
        # Fixme *形参 传入会出错，参考这里 https://typer.tiangolo.com/tutorial/multiple-values/arguments-with-multiple-values/
        typer_commands()



from attrs import define


@define
class Config:
    host: str = '0.0.0.0'
    port: int = 50001
    authkey: bytes = b'sparrow_authkey'


if __name__ == "__main__":
    config = Config(port=2222)
    print(config)



import re


def find_all_index(pattern, string, flags=0):
    """find all matched index of string"""
    return [i.span() for i in re.finditer(pattern, string, flags=flags)]


def string_add(string: str, dx=1):
    # count_points = string.count('.')
    items = find_all_index(r"\.", string)
    number_list = [i for i in string.split(".")]
    number_str = "".join(number_list)
    number_len = len(number_str)
    number = int(number_str)
    number += dx
    new_number_str = f"{number:0>{number_len}d}"
    new_number_list = list(new_number_str)
    [new_number_list.insert(idx[0], ".") for idx in items]
    return "".join(new_number_list)


def index_char(L=1000):
    """
    Get the index of all characters.
    use chr()
    """
    index_token = {}
    token = []
    for i in range(L):
        character = chr(i)
        index_token[i] = character
        token.append(character)
    token_index = dict(zip(token, range(L)))  # token_index[idx] equal to ord(idx)
    return index_token, token_index


def find_match(start, end, S, flag=0):
    """find the string between `start` and `end` of `S`
    flag=0 defaults, means no special specification
    flag options:
        re.I    IGNORECASE， 忽略大小写的匹配模式
        re.M    MULTILINE，多行模式, 改变 ^ 和 $ 的行为
        re.S  　DOTALL，此模式下 '.' 的匹配不受限制，可匹配任何字符，包括换行符，也就是默认是不能匹配换行符
        re.X    VERBOSE，冗余模式， 此模式忽略正则表达式中的空白和#号的注释
    """
    try:
        START = re.search(start, S, flags=flag).span()[1]
        END = re.search(end, S, flags=flag).span()[0]
        return S[START:END]
    except:
        print("Do not match anything.")
        return None


def find_match2(pattern, S, flag=0):
    res = re.search(pattern, S, flags=flag)
    return res.group()



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

import einops
from einops import rearrange, reduce, asnumpy, parse_shape
from einops.layers.torch import Rearrange, Reduce
from torch import einsum


def check_shape(tensor, pattern, **kwargs):
    return einops.rearrange(tensor, f"{pattern} -> {pattern}", **kwargs)


class ScaledDotProductAttention(nn.Module):
    def __init__(self, d_out, d_k, d_v, h, dropout=0.1):
        """
        :param d_out: output dimension
        :param d_k: dimension of key
        :param d_v: dimension of value
        :param h: number of heads
        :param dropout: dropout rate
        """
        super(ScaledDotProductAttention, self).__init__()
        self.d_out = d_out
        self.d_k = d_k
        self.d_v = d_v
        self.h = h
        self.dropout = nn.Dropout(dropout)

        self.fc_q = nn.Linear(d_out, h * d_k)
        self.fc_k = nn.Linear(d_out, h * d_k)
        self.fc_v = nn.Linear(d_out, h * d_v)
        self.fc_o = nn.Linear(h * d_v, d_out)

        self.init_weights()

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
                nn.init.constant_(m.bias, 0.0)
            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.weight, 1.0)
                nn.init.constant_(m.bias, 0.0)
            elif isinstance(m, nn.Dropout):
                nn.init.constant_(m.weight, 1.0)
                nn.init.constant_(m.bias, 0.0)
            elif isinstance(m, nn.Conv2d):
                nn.init.xavier_uniform_(m.weight)
                nn.init.constant_(m.bias, 0.0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out')
                nn.init.constant_(m.bias, 0.0)

    def forward(self, queries, keys, values, mask=None, weights=None):
        """
        :param queries: (batch_size, nq, d_out)
        :param keys: (batch_size, nk, d_out)
        :param values: (batch_size, nk, d_out)
        :param mask: (batch_size, h, nq, nk)
        :param weights: (batch_size, h, nq, nk)
        """
        bs, nq = queries.shape[:2]
        nk = keys.shape[1]

        q = self.fc_q(queries)


def attention_origin(q, k, v):
    """
    attention(q, k, v) = softmax( (q  k.T) / sqrt(dk) )  v
    """
    # for pytorch
    s = q.shape[-1] ** -0.5
    return (q @ k.t() * s).softmax(dim=-1) @ v


def attention(K, V, Q):
    _, n_channels, _ = K.shape
    A = torch.einsum('bct,bcl->btl', [K, Q])
    A = F.softmax(A * n_channels ** (-0.5), 1)
    R = torch.einsum('bct,btl->bcl', [V, A])
    return torch.cat((R, Q), dim=1)


class MultiHeadAttentionNew(nn.Module):
    def __init__(self, n_head, d_model, d_k, d_v, dropout=0.1):
        super().__init__()
        self.n_head = n_head

        self.w_qs = nn.Linear(d_model, n_head * d_k)
        self.w_ks = nn.Linear(d_model, n_head * d_k)
        self.w_vs = nn.Linear(d_model, n_head * d_v)

        nn.init.normal_(self.w_qs.weight, mean=0, std=np.sqrt(2.0 / (d_model + d_k)))
        nn.init.normal_(self.w_ks.weight, mean=0, std=np.sqrt(2.0 / (d_model + d_k)))
        nn.init.normal_(self.w_vs.weight, mean=0, std=np.sqrt(2.0 / (d_model + d_v)))

        self.fc = nn.Linear(n_head * d_v, d_model)
        nn.init.xavier_normal_(self.fc.weight)
        self.dropout = nn.Dropout(p=dropout)
        self.layer_norm = nn.LayerNorm(d_model)

    def forward(self, q, k, v, mask=None):
        residual = q
        q = rearrange(self.w_qs(q), 'b l (head k) -> head b l k', head=self.n_head)
        k = rearrange(self.w_ks(k), 'b t (head k) -> head b t k', head=self.n_head)
        v = rearrange(self.w_vs(v), 'b t (head v) -> head b t v', head=self.n_head)
        attn = torch.einsum('hblk,hbtk->hblt', [q, k]) / np.sqrt(q.shape[-1])
        if mask is not None:
            attn = attn.masked_fill(mask[None], -np.inf)
        attn = torch.softmax(attn, dim=3)
        output = torch.einsum('hblt,hbtv->hblv', [attn, v])
        output = rearrange(output, 'head b l v -> b l (head v)')
        output = self.dropout(self.fc(output))
        output = self.layer_norm(output + residual)
        return output, attn


class Self_Attn_GAN(nn.Module):
    """ Self attention Layer"""

    def __init__(self, in_dim):
        super().__init__()
        self.query_conv = nn.Conv2d(in_dim, out_channels=in_dim // 8, kernel_size=1)
        self.key_conv = nn.Conv2d(in_dim, out_channels=in_dim // 8, kernel_size=1)
        self.value_conv = nn.Conv2d(in_dim, out_channels=in_dim, kernel_size=1)
        self.gamma = nn.Parameter(torch.zeros([1]))

    def forward(self, x):
        proj_query = rearrange(self.query_conv(x), 'b c h w -> b (h w) c')
        proj_key = rearrange(self.key_conv(x), 'b c h w -> b c (h w)')
        proj_value = rearrange(self.value_conv(x), 'b c h w -> b (h w) c')
        energy = torch.bmm(proj_query, proj_key)
        attention = F.softmax(energy, dim=2)
        out = torch.bmm(attention, proj_value)
        out = x + self.gamma * rearrange(out, 'b (h w) c -> b c h w',
                                         **parse_shape(x, 'b c h w'))
        return out, attention


class KroneckerSelfAttention(nn.Module):
    def __init__(self, dim, heads, dim_heads=32):
        super().__init__()
        hidden_dim = heads * dim_heads

        self.heads = heads
        self.to_qkv = nn.Conv1d(dim, hidden_dim * 3, 1, bias=False)
        self.to_out = nn.Conv1d(hidden_dim, dim, 1)

    def forward(self, x):
        h = x.shape[-2]

        x = torch.cat((x.mean(dim=-1), x.mean(dim=-2)), dim=-1)

        qkv = self.to_qkv(x)
        q, k, v = rearrange(qkv, 'b (qkv h d) n -> qkv b h d n', h=self.heads, qkv=3)

        dots = einsum('bhdi,bhdj->bhij', q, k)
        attn = dots.softmax(dim=-1)
        out = einsum('bhij,bhdj->bhdi', attn, v)

        out = rearrange(out, 'b h d n -> b (h d) n')
        out = self.to_out(out)

        # outer sum
        out = rearrange(out[..., :h], 'b c (n 1) -> b c n 1') + rearrange(out[..., h:], 'b c (1 n) -> b c 1 n')
        return out



BLUE_E = "#1C758A"
BLUE_D = "#29ABCA"
BLUE_C = "#58C4DD"
BLUE_B = "#9CDCEB"
BLUE_A = "#C7E9F1"
TEAL_E = "#49A88F"
TEAL_D = "#55C1A7"
TEAL_C = "#5CD0B3"
TEAL_B = "#76DDC0"
TEAL_A = "#ACEAD7"
GREEN_E = "#699C52"
GREEN_D = "#77B05D"
GREEN_C = "#83C167"
GREEN_B = "#A6CF8C"
GREEN_A = "#C9E2AE"
YELLOW_E = "#E8C11C"
YELLOW_D = "#F4D345"
YELLOW_C = "#FFFF00"
YELLOW_B = "#FFEA94"
YELLOW_A = "#FFF1B6"
GOLD_E = "#C78D46"
GOLD_D = "#E1A158"
GOLD_C = "#F0AC5F"
GOLD_B = "#F9B775"
GOLD_A = "#F7C797"
RED_E = "#CF5044"
RED_D = "#E65A4C"
RED_C = "#FC6255"
RED_B = "#FF8080"
RED_A = "#F7A1A3"
MAROON_E = "#94424F"
MAROON_D = "#A24D61"
MAROON_C = "#C55F73"
MAROON_B = "#EC92AB"
MAROON_A = "#ECABC1"
PURPLE_E = "#644172"
PURPLE_D = "#715582"
PURPLE_C = "#9A72AC"
PURPLE_B = "#B189C6"
PURPLE_A = "#CAA3E8"
GREY_E = "#222222"
GREY_D = "#444444"
GREY_C = "#888888"
GREY_B = "#BBBBBB"
GREY_A = "#DDDDDD"
WHITE = "#FFFFFF"
BLACK = "#000000"
GREY_BROWN = "#736357"
DARK_BROWN = "#8B4513"
LIGHT_BROWN = "#CD853F"
PINK = "#D147BD"
LIGHT_PINK = "#DC75CD"
GREEN_SCREEN = "#00FF00"
ORANGE = "#FF862F"

# Abbreviated names for the "median" colors
BLUE = BLUE_C
TEAL = TEAL_C
GREEN = GREEN_C
YELLOW = YELLOW_C
GOLD = GOLD_C
RED = RED_C
MAROON = MAROON_C
PURPLE = PURPLE_C
GREY = GREY_C

# ---------------------------
red_lighter = "#FFEBEE"
red_light = "#EF9A9A"
red = "#EF5350"
red_dark = "#E53935"
red_darker = "#C62828"

pink_lighter = "#FCE4EC"
pink_light = "#F48FB1"
pink = "#EC407A"
pink_dark = "#D81B60"
pink_darker = "#AD1457"

purple_lighter = "#F3E5F5"
purple_light = "#CE93D8"
purple = "#AB47BC"
purple_dark = "#8E24AA"
purple_darker = "#6A1B9A"

deep_purple_lighter = "#EDE7F6"
deep_purple_light = "#B39DDB"
deep_purple = "#7E57C2"
deep_purple_dark = "#5E35B1"
deep_purple_darker = "#4527A0"

indigo_lighter = "#E8EAF6"
indigo_light = "#9FA8DA"
indigo = "#5C6BC0"
indigo_dark = "#3949AB"
indigo_darker = "#283593"

blue_lighter = "#E3F2FD"
blue_light = "#90CAF9"
blue = "#42A5F5"
blue_dark = "#1E88E5"
blue_darker = "#1565C0"

light_blue_lighter = "#E1F5FE"
light_blue_light = "#81D4FA"
light_blue = "#29B6F6"
light_blue_dark = "#039BE5"
light_blue_darker = "#0277BD"

cyan_lighter = "#E0F7FA"
cyan_light = "#80DEEA"
cyan = "#26C6DA"
cyan_dark = "#00ACC1"
cyan_darker = "#00838F"

teal_lighter = "#E0F2F1"
teal_light = "#80CBC4"
teal = "#26A69A"
teal_dark = "#00897B"
teal_darker = "#00695C"

green_lighter = "#E8F5E9"
green_light = "#A5D6A7"
green = "#66BB6A"
green_dark = "#43A047"
green_darker = "#2E7D32"

light_green_lighter = "#F1F8E9"
light_green_light = "#C5E1A5"
light_green = "#9CCC65"
light_green_dark = "#7CB342"
light_green_darker = "#558B2F"

lime_lighter = "#F9FBE7"
lime_light = "#E6EE9C"
lime = "#D4E157"
lime_dark = "#C0CA33"
lime_darker = "#9E9D24"

yellow_lighter = "#FFFDE7"
yellow_light = "#FFF59D"
yellow = "#FFEE58"
yellow_dark = "#FDD835"
yellow_darker = "#F9A825"

amber_lighter = "#FFF8E1"
amber_light = "#FFE082"
amber = "#FFCA28"
amber_dark = "#FFB300"
amber_darker = "#FF8F00"

orange_lighter = "#FFF3E0"
orange_light = "#FFCC80"
orange = "#FFA726"
orange_dark = "#FB8C00"
orange_darker = "#EF6C00"

salmon_lighter = "#FBE9E7"
salmon_light = "#FFAB91"
salmon = "#FF7043"
salmon_dark = "#F4511E"
salmon_darker = "#D84315"

brown_lighter = "#EFEBE9"
brown_light = "#BCAAA4"
brown = "#8D6E63"
brown_dark = "#6D4C41"
brown_darker = "#4E342E"

grey_lighter = "#FAFAFA"
grey_light = "#EEEEEE"
grey = "#BDBDBD"
grey_dark = "#757575"
grey_darker = "#424242"

blue_grey_lighter = "#ECEFF1"
blue_grey_light = "#B0BEC5"
blue_grey = "#78909C"
blue_grey_dark = "#546E7A"
blue_grey_darker = "#37474F"

black = "#000000"
white = "#FFFFFF"



from faker import Faker
from pandas import DataFrame


class Generator:
    def __init__(self, locale="zh_CN"):
        # locale = ['zh_CN', 'en_US', 'ja_JP']
        self._fake = Faker(locale)
        self._dataset = []

    def generate_batch(self, batch_size=10, seed=None):
        Faker.seed(seed)
        name, sentence, address = [], [], []
        phone_nubmer, date = [], []
        free_email, job, company = [], [], []
        for i in range(batch_size):
            name.append(self._fake.name())
            address.append(self._fake.address())
            sentence.append(self._fake.sentence())
            phone_nubmer.append(self._fake.phone_number())
            date.append(self._fake.date())
            free_email.append(self._fake.ascii_free_email())
            job.append(self._fake.job()),
            company.append(self._fake.company())

        return DataFrame(
            {
                "date": date,
                "name": name,
                "address": address,
                "phone_number": phone_nubmer,
                "email": free_email,
                "company": company,
                "job": job,
                # 'sentence': sentence,
            }
        )


if __name__ == "__main__":
    generagor = Generator()
    print(generagor.generate_batch())
    print(generagor.generate_batch(10, seed=0))



from ..color import color_to_int_rgb, color_const

CSI = "\033["
OSC = "\033]"
OFF = CSI + "0m"


def __set_rgb(RGB_fore=(240, 85, 85), SRG=0, RGB_back=None):
    """Get foreground or background color chars
    see https://my.oschina.net/dingdayu/blog/1537064
    inputs:
        RGB_fore: rgb list or tupe of foreground, e.g. [255, 0, 0]
        SRG:  the style of font
        SRG options: see https://en.wikipedia.org/wiki/ANSI_escape_code#SGR
            | 0  |     Close all formats and revert to the original state
            | 1  |     Bold (increased intensity)
            | 2  |     Faint (decreased intensity)
            | 3  |     Italics
            | 4  |     Underline (single line)
            | 5  |     Slow Blink
            | 6  |     Rapid Blink
            | 7  |     Swap the background color with the foreground color
    """
    fore_color = f"{CSI}{SRG};38;2;{RGB_fore[0]};{RGB_fore[1]};{RGB_fore[2]}m"
    if RGB_back is None:
        back_color = ""
    else:
        back_color = f"{CSI}{SRG};48;2;{RGB_back[0]};{RGB_back[1]};{RGB_back[2]}m"
    return fore_color + back_color


def _rgb_str(string, RGB_fore=(240, 85, 85), SRG=0, RGB_back=None):
    return __set_rgb(RGB_fore, SRG, RGB_back) + string + OFF


def rgb_string(string, color=color_const.RED, **kwargs):
    """Return the string with color.
    :param string: The string will be colored.
    :param color: can be rgb list [255, 255, 255]  or hex string "#ffffff".
    :param `SRG`, `RGB_back` see function `__set_grb()`
    :return Colored string.
    """
    rgb = color_to_int_rgb(color)
    return _rgb_str(string, rgb, **kwargs)



import pendulum


class Beijing:
    def __init__(self):
        self.timezones = pendulum.timezones

    @property
    def now(self):
        return pendulum.now("Asia/Shanghai").strftime("%m-%d %H:%M:%S")



import os
from sparrow import rel_to_abs


def milvus(flag='start'):
    db_path = rel_to_abs('./milvus_db')
    if flag == "start":
        os.system(f"cd {db_path} &&make start")
    elif flag == "stop":
        os.system(f"cd {db_path} &&make stop")
    elif flag == 'rm':
        os.system(f"cd {db_path} &&make rm-data")
    else:
        raise ValueError("flag must be 'start' or 'stop' or 'rm'")




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: trainstatus.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11trainstatus.proto\x12\x0bprotostatus\"m\n\x0bTrainStatus\x12\x10\n\x08\x66inished\x18\x01 \x01(\x08\x12\x0c\n\x04step\x18\x02 \x01(\x05\x12\x0c\n\x04loss\x18\x03 \x01(\x02\x12\x10\n\x08progress\x18\x04 \x01(\x02\x12\x11\n\tcost_time\x18\x05 \x01(\x01\x12\x0b\n\x03msg\x18\x06 \x01(\t\"Q\n\x0cHttpResponse\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0b\n\x03msg\x18\x02 \x01(\t\x12&\n\x04\x64\x61ta\x18\x03 \x01(\x0b\x32\x18.protostatus.TrainStatusb\x06proto3')



_TRAINSTATUS = DESCRIPTOR.message_types_by_name['TrainStatus']
_HTTPRESPONSE = DESCRIPTOR.message_types_by_name['HttpResponse']
TrainStatus = _reflection.GeneratedProtocolMessageType('TrainStatus', (_message.Message,), {
  'DESCRIPTOR' : _TRAINSTATUS,
  '__module__' : 'trainstatus_pb2'
  # @@protoc_insertion_point(class_scope:protostatus.TrainStatus)
  })
_sym_db.RegisterMessage(TrainStatus)

HttpResponse = _reflection.GeneratedProtocolMessageType('HttpResponse', (_message.Message,), {
  'DESCRIPTOR' : _HTTPRESPONSE,
  '__module__' : 'trainstatus_pb2'
  # @@protoc_insertion_point(class_scope:protostatus.HttpResponse)
  })
_sym_db.RegisterMessage(HttpResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TRAINSTATUS._serialized_start=34
  _TRAINSTATUS._serialized_end=143
  _HTTPRESPONSE._serialized_start=145
  _HTTPRESPONSE._serialized_end=226
# @@protoc_insertion_point(module_scope)



import numpy as np
import Levenshtein  # pip install python-Levenshtein

"""
Some common distance function
"""


def euclidean_dist(vec1, vec2):
    assert vec1.shape == vec2.shape
    return np.sqrt(np.sum((vec1 - vec2) ** 2))


def manhattan_dist(vec1, vec2):
    return np.sum(np.abs(vec1 - vec2))


def chebyshev_dist(vec1, vec2):
    return np.max(np.abs(vec1 - vec2))


def minkowski_dist(vec1, vec2, p=2):
    """
    :param: `p` The meaning of norm.
        p=1: dist = manhattan_dist
        p=2: dist = euclidean_dist
        p=inf: dist = chebyshev_dist
    """
    s = np.sum(np.power(vec2 - vec1, p))
    return np.power(s, 1 / p)


def cosine_dist(vec1, vec2, p=2):
    # np.linalg.norm(vec, ord=1) 计算p=1范数,默认p=2
    return (vec1.T @ vec2) / (np.linalg.norm(vec1, ord=p) * np.linalg.norm(vec2, ord=p))


def distance(a, b):
    return Levenshtein.ratio(a, b)


def hamming(x, y):
    return np.sum(x != y) / len(x)


def jaccard():
    pass



import shutil
from shutil import _find_unpack_format, _UNPACK_FORMATS
from pathlib import Path
from sparrow.path import rel_to_abs, rel_path_join
import os


def pack(source_path: str, target_path=None, format='gztar'):
    """Pack or compress files.

    Parameters
    ----------
    source_path: str:
        source path
    target_path: str:
        target path
    format : str
        `format` is the archive format: one of "zip", "tar", "gztar"(default), "bztar", or "xztar".
        Or any other registered format.
    """
    if target_path is None:
        target_path = rel_to_abs(source_path, return_str=False, strict=True).name
    new_path = shutil.make_archive(target_path, format, root_dir=source_path)
    print(f"target path:\n{new_path}")


def unpack(filename: str, extract_dir=None, format=None):
    """Unpack or decompress files.

    Parameters
    ----------
    filename : str
        input source file
    extract_dir : str | None
        output dir
    format : str | None
        `format`is the archive format: one of "zip", "tar", "gztar", "bztar", or "xztar".
        If not provided, unpack_archive will use the filename extension.
    """
    name_path = Path(filename)
    if not name_path.exists():
        raise FileExistsError(f"{name_path} not exist.")
    name = name_path.name
    file_format = _find_unpack_format(filename)
    file_postfix_list = _UNPACK_FORMATS[file_format][0]
    for postfix in file_postfix_list:
        if name.endswith(postfix):
            target_name = name[:-len(postfix)]
            break
    else:
        target_name = name.replace('.', '_')

    if extract_dir is None:
        extract_dir = f"./{target_name}/"
    if not Path(extract_dir).exists():
        os.mkdir(extract_dir)
    shutil.unpack_archive(filename, extract_dir, format=format)
    print(f"extract dir:\n{Path(extract_dir).absolute()}")


if __name__ == "__main__":
    # source_path = rel_to_abs('../web/', return_str=False, strict=True).name
    # pack('../web/', format='bztar')
    unpack('./web.tar.bz2')





import os
from sparrow.cli.core import command
from rich import print


def install_node(version='16.x'):
    args = f"curl -sL https://deb.nodesource.com/setup_{version} -o /tmp/nodesource_setup.sh"
    os.system(args)
    os.system("sudo bash /tmp/nodesource_setup.sh")
    print("Update nodesource success, please exec command `apt install nodejs -y` to install Node.js v{version}")


def install_neovim(version="0.8.0"):
    args = f"RUN wget https://github.com/neovim/neovim/releases/download/v{version}/nvim-linux64.deb && " \
           f"sudo apt install ./nvim-linux64.deb"
    os.system(args)
    print(f"Install neovim {version} success.")



from pymilvus import connections, FieldSchema, CollectionSchema, DataType, Collection, utility
import torch
from PIL import Image as PILImage
import cv2
import towhee
from image_utils import get_model, preprocess

dim = 2560
torch_model = get_model()


def efficientnet_b7(img):
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    img = PILImage.fromarray(img.astype('uint8'), 'RGB')
    img = torch.unsqueeze(preprocess(img), 0)
    img = img.to('cuda' if torch.cuda.is_available() else 'cpu')
    embedding = torch_model(img).detach().cpu().numpy()
    return embedding.reshape([dim])


def get_milvus_collection(path_list, collection_name, rebuild=False):
    id_path_list = [(idx, path) for idx, path in enumerate(path_list)]
    connections.connect(host='127.0.0.1', port='19530')
    if rebuild:
        utility.drop_collection(collection_name)
    else:
        if utility.has_collection(collection_name):
            return Collection(collection_name), dict(id_path_list)
        else:
            print("Collection not exist, rebuild it.")

    fields = [
        FieldSchema(name='id',
                    dtype=DataType.INT64,
                    descrition='ids',
                    is_primary=True,
                    auto_id=False),
        # FieldSchema(name='path', dtype=DataType.VARCHAR, descrition='pathes'),
        FieldSchema(name='embedding',
                    dtype=DataType.FLOAT_VECTOR,
                    descrition='embedding vectors',
                    dim=dim)
    ]
    schema = CollectionSchema(fields=fields, description='deduplication image search')
    collection = Collection(name=collection_name, schema=schema)

    # create IVF_FLAT index for collection.
    index_params = {
        'metric_type': 'L2',
        'index_type': "IVF_FLAT",
        'params': {"nlist": 2048}
    }
    collection.create_index(field_name="embedding", index_params=index_params)

    (
        towhee.dc['id', 'path'](id_path_list)
        .runas_op['id', 'id'](func=lambda x: int(x))
        .image_decode['path', 'img']()
        .runas_op['img', 'vec'](func=efficientnet_b7)
        .tensor_normalize['vec', 'vec']()
        .to_milvus['id', 'vec'](collection=collection, batch=100)
    )
    return collection, dict(id_path_list)



import os
from glob import glob
# from pyprobar import probar
from sparrow import probar
from typing import List


def save_docker_images(filedir='.', skip_exists=True, use_stream=False):
    import docker
    from docker.models.images import Image
    client = docker.from_env()
    images_list: List[Image] = client.images.list()
    exist_image_ids = []
    for i in glob(os.path.join(filedir, "*")):
        prefix, filename = os.path.split(i)
        token_list = filename.split(',')
        if len(token_list) > 1:
            exist_image_ids.append(token_list[-1])
    for image in probar(images_list):
        image: Image
        if image.tags:
            if image.id.split(':')[-1] in exist_image_ids and skip_exists:
                print(f"\n image: {image.id} exists, skipping")
            else:
                save_simgle_image(image, filedir, use_stream)


def save_simgle_image(image, filedir='.', use_stream=False):
    image_id = image.id.split(':')[-1]
    filename = f"{image.tags[0].replace('/', '#').replace(':', '@')},{image_id}"
    filepath = os.path.join(filedir, filename)

    if use_stream:
        print(f"\n saving image [{image.tags[0]}] to:", filepath)
        with open(filepath, 'wb') as f:
            for chunk in image.save(named=True):
                f.write(chunk)
    else:
        filepath += ".gz"
        print(f"\n saving image [{image.tags[0]}] to:", filepath)
        image_name = image.tags[0]
        os.system(f"docker save {image_name} | gzip > {filepath}")


def add_tag_to_files(filedir='.'):
    from docker.models.images import Image
    import docker
    client = docker.from_env()
    images_list: List[Image] = client.images.list()
    image_ids_map = dict([[image.id.split(':')[1], str(idx)] for idx, image in enumerate(images_list)])
    file_image_ids = [[i.split('/')[-1], i] for i in glob(os.path.join(filedir, "*"))]
    for file_image_id, filename in file_image_ids:
        str_image_idx = image_ids_map.get(file_image_id)
        if str_image_idx:
            image = images_list[int(str_image_idx)]
            prefix, file_id_name = os.path.split(filename)
            os.rename(filename,
                      str(os.path.join(prefix, f"{image.tags[0].replace('/', '#').replace(':', '@')},{file_id_name}")))


def load_docker_images(filename_pattern="./*", skip_exists=True):
    from docker.models.images import Image
    import docker
    client = docker.from_env()
    images_list: List[Image] = client.images.list()
    exist_image_ids = [image.id.split(":")[-1] for image in images_list]

    for filename in probar(glob(filename_pattern)):
        filename: str
        file_name = os.path.split(filename)[-1]
        file_image_id = file_name.split(',')[1]
        if file_image_id in exist_image_ids and skip_exists:
            print(f"\n image id: {file_image_id} exists, skipping")
        else:
            if filename.endswith('.gz'):
                os.system(f"gunzip -c {filename}| docker load")
            else:
                os.system(f"docker load -i {filename}")


def find_by_pid(pid: int):
    command = f"""\
dockerid=`cat /proc/{pid}/cgroup | grep -oPm1 .*/docker/.* | sed -e 's/.*docker\///'`
shortid=`echo $dockerid |cut -c1-12`
echo $shortid
docker ps |grep $shortid
"""
    os.system(command)


if __name__ == "__main__":
    filedir = "/media/kunyuan/系统/popos_docker"
    save_docker_images(filedir, use_stream=False)
    # load_dir_images(f"{filedir}/*")



imap = lambda *args, **kwargs: list(map(*args, **kwargs))
ifilter = lambda *args, **kwargs: list(filter(*args, **kwargs))


class BKTree:
    def __init__(self, distfn, words):
        """
        Create a new BK-tree from the given distance function and
        words.

        Arguments:

        distfn: a binary function that returns the distance between
        two words.  Return value is a non-negative integer.  the
        distance function must be a metric space.

        words: an iterable.  produces values that can be passed to
        distfn

        """
        self.distfn = distfn

        it = iter(words)
        root = it.next()
        self.tree = (root, {})

        for i in it:
            self._add_word(self.tree, i)

    def _add_word(self, parent, word):
        pword, children = parent
        d = self.distfn(word, pword)
        if d in children:
            self._add_word(children[d], word)
        else:
            children[d] = (word, {})

    def query(self, word, n):
        """
        Return all words in the tree that are within a distance of `n'
        from `word`.

        Arguments:

        word: a word to query on

        n: a non-negative integer that specifies the allowed distance
        from the query word.

        Return value is a list of tuples (distance, word), sorted in
        ascending order of distance.

        """

        def rec(parent):
            pword, children = parent
            d = self.distfn(word, pword)
            results = []
            if d <= n:
                results.append((d, pword))

            for i in range(d - n, d + n + 1):
                child = children.get(i)
                if child is not None:
                    results.extend(rec(child))
            return results

        # sort by distance
        return sorted(rec(self.tree))


def brute_query(word, words, distfn, n):
    """A brute force distance query

    Arguments:

    word: the word to query for

    words: a iterable that produces words to test

    distfn: a binary function that returns the distance between a
    `word' and an item in `words'.

    n: an integer that specifies the distance of a matching word

    """
    return [i for i in words
            if distfn(i, word) <= n]


# def maxdepth(tree, count=0):
#     _, children = t
#     if len(children):
#         return max(maxdepth(i, c + 1) for i in children.values())
#     else:
#         return c


# http://en.wikibooks.org/wiki/Algorithm_implementation/Strings/Levenshtein_distance#Python
def levenshtein(s, t):
    m, n = len(s), len(t)
    d = [range(n + 1)]
    d += [[i] for i in range(1, m + 1)]
    for i in range(0, m):
        for j in range(0, n):
            cost = 1
            if s[i] == t[j]: cost = 0

            d[i + 1].append(min(d[i][j + 1] + 1,  # deletion
                                d[i + 1][j] + 1,  # insertion
                                d[i][j] + cost)  # substitution
                            )
    return d[m][n]


if __name__ == "__main__":
    ...



from functools import wraps
import pickle


def save_var(filename, data):
    with open(filename, "wb") as fw:
        pickle.dump(data, fw)


def load_var(filename):
    with open(filename, "rb") as fi:
        data = pickle.load(fi)
    return data


def broadcast(func):  # It can be replaced by `np.vectorize`
    """Only for a functions with a single argument
    example:
    @broadcast
    def f(x):
        # A function that can map only a single element
        if x==1 or x==0:
            return x
        else:
            return f(x-1)+f(x-2)

    >> f([2,4,10])
    >> (1, 3, 832040)
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        value_list = []
        for arg in args:
            value_list.append(func(arg, **kwargs))
        return tuple(value_list)

    return wrapper



from sparrow.utils.core import OrderedSet


_SIMPLE = "万与丑专业丛东丝丢两严丧个丬丰临为丽举么义乌乐乔习乡书买乱争于亏云亘亚产亩亲亵亸亿仅从仑仓仪们价众优伙会伛伞伟传伤伥伦伧伪伫体余佣佥侠侣侥侦侧侨侩侪侬俣俦俨俩俪俭债倾偬偻偾偿傥傧储傩儿兑兖党兰关兴兹养兽冁内冈册写军农冢冯冲决况冻净凄凉凌减凑凛几凤凫凭凯击凼凿刍划刘则刚创删别刬刭刽刿剀剂剐剑剥剧劝办务劢动励劲劳势勋勐勚匀匦匮区医华协单卖卢卤卧卫却卺厂厅历厉压厌厍厕厢厣厦厨厩厮县参叆叇双发变叙叠叶号叹叽吁后吓吕吗吣吨听启吴呒呓呕呖呗员呙呛呜咏咔咙咛咝咤咴咸哌响哑哒哓哔哕哗哙哜哝哟唛唝唠唡唢唣唤唿啧啬啭啮啰啴啸喷喽喾嗫呵嗳嘘嘤嘱噜噼嚣嚯团园囱围囵国图圆圣圹场坂坏块坚坛坜坝坞坟坠垄垅垆垒垦垧垩垫垭垯垱垲垴埘埙埚埝埯堑堕塆墙壮声壳壶壸处备复够头夸夹夺奁奂奋奖奥妆妇妈妩妪妫姗姜娄娅娆娇娈娱娲娴婳婴婵婶媪嫒嫔嫱嬷孙学孪宁宝实宠审宪宫宽宾寝对寻导寿将尔尘尧尴尸尽层屃屉届属屡屦屿岁岂岖岗岘岙岚岛岭岳岽岿峃峄峡峣峤峥峦崂崃崄崭嵘嵚嵛嵝嵴巅巩巯币帅师帏帐帘帜带帧帮帱帻帼幂幞干并广庄庆庐庑库应庙庞废庼廪开异弃张弥弪弯弹强归当录彟彦彻径徕御忆忏忧忾怀态怂怃怄怅怆怜总怼怿恋恳恶恸恹恺恻恼恽悦悫悬悭悯惊惧惨惩惫惬惭惮惯愍愠愤愦愿慑慭憷懑懒懔戆戋戏戗战戬户扎扑扦执扩扪扫扬扰抚抛抟抠抡抢护报担拟拢拣拥拦拧拨择挂挚挛挜挝挞挟挠挡挢挣挤挥挦捞损捡换捣据捻掳掴掷掸掺掼揸揽揿搀搁搂搅携摄摅摆摇摈摊撄撑撵撷撸撺擞攒敌敛数斋斓斗斩断无旧时旷旸昙昼昽显晋晒晓晔晕晖暂暧札术朴机杀杂权条来杨杩杰极构枞枢枣枥枧枨枪枫枭柜柠柽栀栅标栈栉栊栋栌栎栏树栖样栾桊桠桡桢档桤桥桦桧桨桩梦梼梾检棂椁椟椠椤椭楼榄榇榈榉槚槛槟槠横樯樱橥橱橹橼檐檩欢欤欧歼殁殇残殒殓殚殡殴毁毂毕毙毡毵氇气氢氩氲汇汉污汤汹沓沟没沣沤沥沦沧沨沩沪沵泞泪泶泷泸泺泻泼泽泾洁洒洼浃浅浆浇浈浉浊测浍济浏浐浑浒浓浔浕涂涌涛涝涞涟涠涡涢涣涤润涧涨涩淀渊渌渍渎渐渑渔渖渗温游湾湿溃溅溆溇滗滚滞滟滠满滢滤滥滦滨滩滪漤潆潇潋潍潜潴澜濑濒灏灭灯灵灾灿炀炉炖炜炝点炼炽烁烂烃烛烟烦烧烨烩烫烬热焕焖焘煅煳熘爱爷牍牦牵牺犊犟状犷犸犹狈狍狝狞独狭狮狯狰狱狲猃猎猕猡猪猫猬献獭玑玙玚玛玮环现玱玺珉珏珐珑珰珲琎琏琐琼瑶瑷璇璎瓒瓮瓯电画畅畲畴疖疗疟疠疡疬疮疯疱疴痈痉痒痖痨痪痫痴瘅瘆瘗瘘瘪瘫瘾瘿癞癣癫癯皑皱皲盏盐监盖盗盘眍眦眬着睁睐睑瞒瞩矫矶矾矿砀码砖砗砚砜砺砻砾础硁硅硕硖硗硙硚确硷碍碛碜碱碹磙礼祎祢祯祷祸禀禄禅离秃秆种积称秽秾稆税稣稳穑穷窃窍窑窜窝窥窦窭竖竞笃笋笔笕笺笼笾筑筚筛筜筝筹签简箓箦箧箨箩箪箫篑篓篮篱簖籁籴类籼粜粝粤粪粮糁糇紧絷纟纠纡红纣纤纥约级纨纩纪纫纬纭纮纯纰纱纲纳纴纵纶纷纸纹纺纻纼纽纾线绀绁绂练组绅细织终绉绊绋绌绍绎经绐绑绒结绔绕绖绗绘给绚绛络绝绞统绠绡绢绣绤绥绦继绨绩绪绫绬续绮绯绰绱绲绳维绵绶绷绸绹绺绻综绽绾绿缀缁缂缃缄缅缆缇缈缉缊缋缌缍缎缏缐缑缒缓缔缕编缗缘缙缚缛缜缝缞缟缠缡缢缣缤缥缦缧缨缩缪缫缬缭缮缯缰缱缲缳缴缵罂网罗罚罢罴羁羟羡翘翙翚耢耧耸耻聂聋职聍联聩聪肃肠肤肷肾肿胀胁胆胜胧胨胪胫胶脉脍脏脐脑脓脔脚脱脶脸腊腌腘腭腻腼腽腾膑臜舆舣舰舱舻艰艳艹艺节芈芗芜芦苁苇苈苋苌苍苎苏苘苹茎茏茑茔茕茧荆荐荙荚荛荜荞荟荠荡荣荤荥荦荧荨荩荪荫荬荭荮药莅莜莱莲莳莴莶获莸莹莺莼萚萝萤营萦萧萨葱蒇蒉蒋蒌蓝蓟蓠蓣蓥蓦蔷蔹蔺蔼蕲蕴薮藁藓虏虑虚虫虬虮虽虾虿蚀蚁蚂蚕蚝蚬蛊蛎蛏蛮蛰蛱蛲蛳蛴蜕蜗蜡蝇蝈蝉蝎蝼蝾螀螨蟏衅衔补衬衮袄袅袆袜袭袯装裆裈裢裣裤裥褛褴襁襕见观觃规觅视觇览觉觊觋觌觍觎觏觐觑觞触觯詟誉誊讠计订讣认讥讦讧讨让讪讫训议讯记讱讲讳讴讵讶讷许讹论讻讼讽设访诀证诂诃评诅识诇诈诉诊诋诌词诎诏诐译诒诓诔试诖诗诘诙诚诛诜话诞诟诠诡询诣诤该详诧诨诩诪诫诬语诮误诰诱诲诳说诵诶请诸诹诺读诼诽课诿谀谁谂调谄谅谆谇谈谊谋谌谍谎谏谐谑谒谓谔谕谖谗谘谙谚谛谜谝谞谟谠谡谢谣谤谥谦谧谨谩谪谫谬谭谮谯谰谱谲谳谴谵谶谷豮贝贞负贠贡财责贤败账货质贩贪贫贬购贮贯贰贱贲贳贴贵贶贷贸费贺贻贼贽贾贿赀赁赂赃资赅赆赇赈赉赊赋赌赍赎赏赐赑赒赓赔赕赖赗赘赙赚赛赜赝赞赟赠赡赢赣赪赵赶趋趱趸跃跄跖跞践跶跷跸跹跻踊踌踪踬踯蹑蹒蹰蹿躏躜躯车轧轨轩轪轫转轭轮软轰轱轲轳轴轵轶轷轸轹轺轻轼载轾轿辀辁辂较辄辅辆辇辈辉辊辋辌辍辎辏辐辑辒输辔辕辖辗辘辙辚辞辩辫边辽达迁过迈运还这进远违连迟迩迳迹适选逊递逦逻遗遥邓邝邬邮邹邺邻郁郄郏郐郑郓郦郧郸酝酦酱酽酾酿释里鉅鉴銮錾钆钇针钉钊钋钌钍钎钏钐钑钒钓钔钕钖钗钘钙钚钛钝钞钟钠钡钢钣钤钥钦钧钨钩钪钫钬钭钮钯钰钱钲钳钴钵钶钷钸钹钺钻钼钽钾钿铀铁铂铃铄铅铆铈铉铊铋铍铎铏铐铑铒铕铗铘铙铚铛铜铝铞铟铠铡铢铣铤铥铦铧铨铪铫铬铭铮铯铰铱铲铳铴铵银铷铸铹铺铻铼铽链铿销锁锂锃锄锅锆锇锈锉锊锋锌锍锎锏锐锑锒锓锔锕锖锗错锚锜锞锟锠锡锢锣锤锥锦锨锩锫锬锭键锯锰锱锲锳锴锵锶锷锸锹锺锻锼锽锾锿镀镁镂镃镆镇镈镉镊镌镍镎镏镐镑镒镕镖镗镙镚镛镜镝镞镟镠镡镢镣镤镥镦镧镨镩镪镫镬镭镮镯镰镱镲镳镴镶长门闩闪闫闬闭问闯闰闱闲闳间闵闶闷闸闹闺闻闼闽闾闿阀阁阂阃阄阅阆阇阈阉阊阋阌阍阎阏阐阑阒阓阔阕阖阗阘阙阚阛队阳阴阵阶际陆陇陈陉陕陧陨险随隐隶隽难雏雠雳雾霁霉霭靓静靥鞑鞒鞯鞴韦韧韨韩韪韫韬韵页顶顷顸项顺须顼顽顾顿颀颁颂颃预颅领颇颈颉颊颋颌颍颎颏颐频颒颓颔颕颖颗题颙颚颛颜额颞颟颠颡颢颣颤颥颦颧风飏飐飑飒飓飔飕飖飗飘飙飚飞飨餍饤饥饦饧饨饩饪饫饬饭饮饯饰饱饲饳饴饵饶饷饸饹饺饻饼饽饾饿馀馁馂馃馄馅馆馇馈馉馊馋馌馍馎馏馐馑馒馓馔馕马驭驮驯驰驱驲驳驴驵驶驷驸驹驺驻驼驽驾驿骀骁骂骃骄骅骆骇骈骉骊骋验骍骎骏骐骑骒骓骔骕骖骗骘骙骚骛骜骝骞骟骠骡骢骣骤骥骦骧髅髋髌鬓魇魉鱼鱽鱾鱿鲀鲁鲂鲄鲅鲆鲇鲈鲉鲊鲋鲌鲍鲎鲏鲐鲑鲒鲓鲔鲕鲖鲗鲘鲙鲚鲛鲜鲝鲞鲟鲠鲡鲢鲣鲤鲥鲦鲧鲨鲩鲪鲫鲬鲭鲮鲯鲰鲱鲲鲳鲴鲵鲶鲷鲸鲹鲺鲻鲼鲽鲾鲿鳀鳁鳂鳃鳄鳅鳆鳇鳈鳉鳊鳋鳌鳍鳎鳏鳐鳑鳒鳓鳔鳕鳖鳗鳘鳙鳛鳜鳝鳞鳟鳠鳡鳢鳣鸟鸠鸡鸢鸣鸤鸥鸦鸧鸨鸩鸪鸫鸬鸭鸮鸯鸰鸱鸲鸳鸴鸵鸶鸷鸸鸹鸺鸻鸼鸽鸾鸿鹀鹁鹂鹃鹄鹅鹆鹇鹈鹉鹊鹋鹌鹍鹎鹏鹐鹑鹒鹓鹔鹕鹖鹗鹘鹚鹛鹜鹝鹞鹟鹠鹡鹢鹣鹤鹥鹦鹧鹨鹩鹪鹫鹬鹭鹯鹰鹱鹲鹳鹴鹾麦麸黄黉黡黩黪黾鼋鼌鼍鼗鼹齄齐齑齿龀龁龂龃龄龅龆龇龈龉龊龋龌龙龚龛龟志制咨只里系范松没尝尝闹面准钟别闲干尽脏拼"
_TRADITION = "萬與醜專業叢東絲丟兩嚴喪個爿豐臨為麗舉麼義烏樂喬習鄉書買亂爭於虧雲亙亞產畝親褻嚲億僅從侖倉儀們價眾優夥會傴傘偉傳傷倀倫傖偽佇體餘傭僉俠侶僥偵側僑儈儕儂俁儔儼倆儷儉債傾傯僂僨償儻儐儲儺兒兌兗黨蘭關興茲養獸囅內岡冊寫軍農塚馮衝決況凍淨淒涼淩減湊凜幾鳳鳧憑凱擊氹鑿芻劃劉則剛創刪別剗剄劊劌剴劑剮劍剝劇勸辦務勱動勵勁勞勢勳猛勩勻匭匱區醫華協單賣盧鹵臥衛卻巹廠廳曆厲壓厭厙廁廂厴廈廚廄廝縣參靉靆雙發變敘疊葉號歎嘰籲後嚇呂嗎唚噸聽啟吳嘸囈嘔嚦唄員咼嗆嗚詠哢嚨嚀噝吒噅鹹呱響啞噠嘵嗶噦嘩噲嚌噥喲嘜嗊嘮啢嗩唕喚呼嘖嗇囀齧囉嘽嘯噴嘍嚳囁嗬噯噓嚶囑嚕劈囂謔團園囪圍圇國圖圓聖壙場阪壞塊堅壇壢壩塢墳墜壟壟壚壘墾坰堊墊埡墶壋塏堖塒塤堝墊垵塹墮壪牆壯聲殼壺壼處備複夠頭誇夾奪奩奐奮獎奧妝婦媽嫵嫗媯姍薑婁婭嬈嬌孌娛媧嫻嫿嬰嬋嬸媼嬡嬪嬙嬤孫學孿寧寶實寵審憲宮寬賓寢對尋導壽將爾塵堯尷屍盡層屭屜屆屬屢屨嶼歲豈嶇崗峴嶴嵐島嶺嶽崠巋嶨嶧峽嶢嶠崢巒嶗崍嶮嶄嶸嶔崳嶁脊巔鞏巰幣帥師幃帳簾幟帶幀幫幬幘幗冪襆幹並廣莊慶廬廡庫應廟龐廢廎廩開異棄張彌弳彎彈強歸當錄彠彥徹徑徠禦憶懺憂愾懷態慫憮慪悵愴憐總懟懌戀懇惡慟懨愷惻惱惲悅愨懸慳憫驚懼慘懲憊愜慚憚慣湣慍憤憒願懾憖怵懣懶懍戇戔戲戧戰戩戶紮撲扡執擴捫掃揚擾撫拋摶摳掄搶護報擔擬攏揀擁攔擰撥擇掛摯攣掗撾撻挾撓擋撟掙擠揮撏撈損撿換搗據撚擄摑擲撣摻摜摣攬撳攙擱摟攪攜攝攄擺搖擯攤攖撐攆擷擼攛擻攢敵斂數齋斕鬥斬斷無舊時曠暘曇晝曨顯晉曬曉曄暈暉暫曖劄術樸機殺雜權條來楊榪傑極構樅樞棗櫪梘棖槍楓梟櫃檸檉梔柵標棧櫛櫳棟櫨櫟欄樹棲樣欒棬椏橈楨檔榿橋樺檜槳樁夢檮棶檢欞槨櫝槧欏橢樓欖櫬櫚櫸檟檻檳櫧橫檣櫻櫫櫥櫓櫞簷檁歡歟歐殲歿殤殘殞殮殫殯毆毀轂畢斃氈毿氌氣氫氬氳彙漢汙湯洶遝溝沒灃漚瀝淪滄渢溈滬濔濘淚澩瀧瀘濼瀉潑澤涇潔灑窪浹淺漿澆湞溮濁測澮濟瀏滻渾滸濃潯濜塗湧濤澇淶漣潿渦溳渙滌潤澗漲澀澱淵淥漬瀆漸澠漁瀋滲溫遊灣濕潰濺漵漊潷滾滯灩灄滿瀅濾濫灤濱灘澦濫瀠瀟瀲濰潛瀦瀾瀨瀕灝滅燈靈災燦煬爐燉煒熗點煉熾爍爛烴燭煙煩燒燁燴燙燼熱煥燜燾煆糊溜愛爺牘犛牽犧犢強狀獷獁猶狽麅獮獰獨狹獅獪猙獄猻獫獵獼玀豬貓蝟獻獺璣璵瑒瑪瑋環現瑲璽瑉玨琺瓏璫琿璡璉瑣瓊瑤璦璿瓔瓚甕甌電畫暢佘疇癤療瘧癘瘍鬁瘡瘋皰屙癰痙癢瘂癆瘓癇癡癉瘮瘞瘺癟癱癮癭癩癬癲臒皚皺皸盞鹽監蓋盜盤瞘眥矓著睜睞瞼瞞矚矯磯礬礦碭碼磚硨硯碸礪礱礫礎硜矽碩硤磽磑礄確鹼礙磧磣堿镟滾禮禕禰禎禱禍稟祿禪離禿稈種積稱穢穠穭稅穌穩穡窮竊竅窯竄窩窺竇窶豎競篤筍筆筧箋籠籩築篳篩簹箏籌簽簡籙簀篋籜籮簞簫簣簍籃籬籪籟糴類秈糶糲粵糞糧糝餱緊縶糸糾紆紅紂纖紇約級紈纊紀紉緯紜紘純紕紗綱納紝縱綸紛紙紋紡紵紖紐紓線紺絏紱練組紳細織終縐絆紼絀紹繹經紿綁絨結絝繞絰絎繪給絢絳絡絕絞統綆綃絹繡綌綏絛繼綈績緒綾緓續綺緋綽緔緄繩維綿綬繃綢綯綹綣綜綻綰綠綴緇緙緗緘緬纜緹緲緝縕繢緦綞緞緶線緱縋緩締縷編緡緣縉縛縟縝縫縗縞纏縭縊縑繽縹縵縲纓縮繆繅纈繚繕繒韁繾繰繯繳纘罌網羅罰罷羆羈羥羨翹翽翬耮耬聳恥聶聾職聹聯聵聰肅腸膚膁腎腫脹脅膽勝朧腖臚脛膠脈膾髒臍腦膿臠腳脫腡臉臘醃膕齶膩靦膃騰臏臢輿艤艦艙艫艱豔艸藝節羋薌蕪蘆蓯葦藶莧萇蒼苧蘇檾蘋莖蘢蔦塋煢繭荊薦薘莢蕘蓽蕎薈薺蕩榮葷滎犖熒蕁藎蓀蔭蕒葒葤藥蒞蓧萊蓮蒔萵薟獲蕕瑩鶯蓴蘀蘿螢營縈蕭薩蔥蕆蕢蔣蔞藍薊蘺蕷鎣驀薔蘞藺藹蘄蘊藪槁蘚虜慮虛蟲虯蟣雖蝦蠆蝕蟻螞蠶蠔蜆蠱蠣蟶蠻蟄蛺蟯螄蠐蛻蝸蠟蠅蟈蟬蠍螻蠑螿蟎蠨釁銜補襯袞襖嫋褘襪襲襏裝襠褌褳襝褲襇褸襤繈襴見觀覎規覓視覘覽覺覬覡覿覥覦覯覲覷觴觸觶讋譽謄訁計訂訃認譏訐訌討讓訕訖訓議訊記訒講諱謳詎訝訥許訛論訩訟諷設訪訣證詁訶評詛識詗詐訴診詆謅詞詘詔詖譯詒誆誄試詿詩詰詼誠誅詵話誕詬詮詭詢詣諍該詳詫諢詡譸誡誣語誚誤誥誘誨誑說誦誒請諸諏諾讀諑誹課諉諛誰諗調諂諒諄誶談誼謀諶諜謊諫諧謔謁謂諤諭諼讒諮諳諺諦謎諞諝謨讜謖謝謠謗諡謙謐謹謾謫譾謬譚譖譙讕譜譎讞譴譫讖穀豶貝貞負貟貢財責賢敗賬貨質販貪貧貶購貯貫貳賤賁貰貼貴貺貸貿費賀貽賊贄賈賄貲賃賂贓資賅贐賕賑賚賒賦賭齎贖賞賜贔賙賡賠賧賴賵贅賻賺賽賾贗讚贇贈贍贏贛赬趙趕趨趲躉躍蹌蹠躒踐躂蹺蹕躚躋踴躊蹤躓躑躡蹣躕躥躪躦軀車軋軌軒軑軔轉軛輪軟轟軲軻轤軸軹軼軤軫轢軺輕軾載輊轎輈輇輅較輒輔輛輦輩輝輥輞輬輟輜輳輻輯轀輸轡轅轄輾轆轍轔辭辯辮邊遼達遷過邁運還這進遠違連遲邇逕跡適選遜遞邐邏遺遙鄧鄺鄔郵鄒鄴鄰鬱郤郟鄶鄭鄆酈鄖鄲醞醱醬釅釃釀釋裏钜鑒鑾鏨釓釔針釘釗釙釕釷釺釧釤鈒釩釣鍆釹鍚釵鈃鈣鈈鈦鈍鈔鍾鈉鋇鋼鈑鈐鑰欽鈞鎢鉤鈧鈁鈥鈄鈕鈀鈺錢鉦鉗鈷缽鈳鉕鈽鈸鉞鑽鉬鉭鉀鈿鈾鐵鉑鈴鑠鉛鉚鈰鉉鉈鉍鈹鐸鉶銬銠鉺銪鋏鋣鐃銍鐺銅鋁銱銦鎧鍘銖銑鋌銩銛鏵銓鉿銚鉻銘錚銫鉸銥鏟銃鐋銨銀銣鑄鐒鋪鋙錸鋱鏈鏗銷鎖鋰鋥鋤鍋鋯鋨鏽銼鋝鋒鋅鋶鐦鐧銳銻鋃鋟鋦錒錆鍺錯錨錡錁錕錩錫錮鑼錘錐錦鍁錈錇錟錠鍵鋸錳錙鍥鍈鍇鏘鍶鍔鍤鍬鍾鍛鎪鍠鍰鎄鍍鎂鏤鎡鏌鎮鎛鎘鑷鐫鎳鎿鎦鎬鎊鎰鎔鏢鏜鏍鏰鏞鏡鏑鏃鏇鏐鐔钁鐐鏷鑥鐓鑭鐠鑹鏹鐙鑊鐳鐶鐲鐮鐿鑔鑣鑞鑲長門閂閃閆閈閉問闖閏闈閑閎間閔閌悶閘鬧閨聞闥閩閭闓閥閣閡閫鬮閱閬闍閾閹閶鬩閿閽閻閼闡闌闃闠闊闋闔闐闒闕闞闤隊陽陰陣階際陸隴陳陘陝隉隕險隨隱隸雋難雛讎靂霧霽黴靄靚靜靨韃鞽韉韝韋韌韍韓韙韞韜韻頁頂頃頇項順須頊頑顧頓頎頒頌頏預顱領頗頸頡頰頲頜潁熲頦頤頻頮頹頷頴穎顆題顒顎顓顏額顳顢顛顙顥纇顫顬顰顴風颺颭颮颯颶颸颼颻飀飄飆飆飛饗饜飣饑飥餳飩餼飪飫飭飯飲餞飾飽飼飿飴餌饒餉餄餎餃餏餅餑餖餓餘餒餕餜餛餡館餷饋餶餿饞饁饃餺餾饈饉饅饊饌饢馬馭馱馴馳驅馹駁驢駔駛駟駙駒騶駐駝駑駕驛駘驍罵駰驕驊駱駭駢驫驪騁驗騂駸駿騏騎騍騅騌驌驂騙騭騤騷騖驁騮騫騸驃騾驄驏驟驥驦驤髏髖髕鬢魘魎魚魛魢魷魨魯魴魺鮁鮃鯰鱸鮋鮓鮒鮊鮑鱟鮍鮐鮭鮚鮳鮪鮞鮦鰂鮜鱠鱭鮫鮮鮺鯗鱘鯁鱺鰱鰹鯉鰣鰷鯀鯊鯇鮶鯽鯒鯖鯪鯕鯫鯡鯤鯧鯝鯢鯰鯛鯨鯵鯴鯔鱝鰈鰏鱨鯷鰮鰃鰓鱷鰍鰒鰉鰁鱂鯿鰠鼇鰭鰨鰥鰩鰟鰜鰳鰾鱈鱉鰻鰵鱅鰼鱖鱔鱗鱒鱯鱤鱧鱣鳥鳩雞鳶鳴鳲鷗鴉鶬鴇鴆鴣鶇鸕鴨鴞鴦鴒鴟鴝鴛鴬鴕鷥鷙鴯鴰鵂鴴鵃鴿鸞鴻鵐鵓鸝鵑鵠鵝鵒鷳鵜鵡鵲鶓鵪鶤鵯鵬鵮鶉鶊鵷鷫鶘鶡鶚鶻鶿鶥鶩鷊鷂鶲鶹鶺鷁鶼鶴鷖鸚鷓鷚鷯鷦鷲鷸鷺鸇鷹鸌鸏鸛鸘鹺麥麩黃黌黶黷黲黽黿鼂鼉鞀鼴齇齊齏齒齔齕齗齟齡齙齠齜齦齬齪齲齷龍龔龕龜誌製谘隻裡係範鬆冇嚐嘗鬨麵準鐘彆閒乾儘臟拚"


# _simple_set = OrderedSet([i for i in _SIMPLE])
# _tradition_set = OrderedSet([i for i in _TRADITION])

# simple to tradition
def to_tradition(str):
    output_str_list = []
    str_len = len(str)
    for i in range(str_len):
        found_index = _SIMPLE.find(str[i])
        if not (found_index == -1):
            output_str_list.append(_TRADITION[found_index])
        else:
            output_str_list.append(str[i])
    return "".join(output_str_list)


# tradition to simple
def to_simple(str):
    output_str_list = []
    str_len = len(str)
    for i in range(str_len):
        found_index = _TRADITION.find(str[i])
        if not (found_index == -1):
            output_str_list.append(_SIMPLE[found_index])
        else:
            output_str_list.append(str[i])
    return "".join(output_str_list)



import numpy as np


def get_continue_subseries(series, idx):
    """
    Examples:
        >>> series = [0, 2, 3, 4, 5, 7, 8, 10, 11]
        >>> get_continue_subseries(series, 2)
        array([2, 3, 4, 5])
        >>> get_continue_subseries(series, -1)
        array([10, 11])
        >>> get_continue_subseries(series, 0)
        array([0])
    """

    idx_min, ref_min = idx, series[idx]
    while ref_min == series[idx_min]:
        idx_min -= 1
        ref_min -= 1

    idx_max, ref_max = idx, series[idx]
    while idx_max < len(series) and ref_max == series[idx_max]:
        idx_max += 1
        ref_max += 1
    idxs = np.arange(idx_min + 1, idx_max)
    return np.asarray(series)[idxs]


def series_concatenate_and_sort(series_a, series_b):
    """
    :returns:
        concatenated and sorted series.
        series_b's index in concatenated series.
    """
    t_series = np.concatenate((series_a, series_b))
    sorted_idxs = np.argsort(t_series)
    series_b_idxs = np.argsort(sorted_idxs)[-len(series_b) :]
    return t_series[sorted_idxs], series_b_idxs


def calc_continue_series_dict(target_series, key_series):
    """
    Examples:
        >>> a = (0, 1, 4, 5, 7, 8, 9, 11, 12, 13, 15, 16, 17)
        >>> b = (2, 10,)
        >>> calc_continue_series_dict(a, b)
        {2: array([0, 1, 2]), 10: array([ 7,  8,  9, 10, 11, 12, 13])}
    """
    concatenated_series, flag_idxs = series_concatenate_and_sort(
        target_series, key_series
    )
    target_dict = {}
    for idx in range(len(key_series)):
        subseries = get_continue_subseries(concatenated_series, flag_idxs[idx])
        target_dict.setdefault(key_series[idx], subseries)
    return target_dict


if __name__ == "__main__":
    a = (0, 1, 4, 5, 7, 8, 9, 11, 12, 13, 15, 16, 17)
    b = (2, 10, 14)
    print(calc_continue_series_dict(a, b))
    # import doctest
    # doctest.testmod()



from .core import choose


def bezier(points):
    n = len(points) - 1

    def result(t):
        return sum(
            [
                ((1 - t) ** (n - k)) * (t**k) * choose(n, k) * point
                for k, point in enumerate(points)
            ]
        )

    return result



import time
from collections import deque
from sparrow.string.color_string import rgb_string
from sparrow.color.constant import TEAL, GREEN


def timer(dt=0.01):
    """A simple timer.
    Press `space` to start and suspend.
    press `q` to quit.
    """
    try:
        import keyboard
    except ImportError as e:
        raise ImportError("import `keyboard` error, use pip to install: `pip install keyboard`")
    print(
        rgb_string("Press `space` to start and suspend.", color=TEAL),
    )
    q = deque(maxlen=1)
    q.append(True)
    keyboard.wait("space")
    keyboard.add_hotkey("space", lambda: q.append(not q[0]))
    t0 = time.time()
    current_time = 0
    while True:
        time.sleep(dt)
        ct = time.time()
        if q[0]:
            t0 = ct
        else:
            current_time += ct - t0
            print(rgb_string(f"\r{current_time:.3f} secs", color=GREEN), end="")
            t0 = ct
        if keyboard.is_pressed("q"):
            break




from colour import Color
import numpy as np
from . import constant as _color_const


def color_to_rgb(color):
    if isinstance(color, str):
        return hex_to_rgb(color)
    elif isinstance(color, Color):
        return np.array(color.get_rgb())
    elif isinstance(color, (tuple, list, np.ndarray)):

        if isinstance(color[0], (int, np.int32, np.int64)):
            return np.array(color) / 255
        elif isinstance(color[0], (float, np.float32, np.float64)):
            return np.array(color)
        else:
            raise Exception(f"Invalid color type: {type(color[0])}")

    else:
        raise Exception(f"Invalid color type: {color}")


def color_to_rgba(color, alpha=1):
    return np.array([*color_to_rgb(color), alpha])


def rgb_to_color(rgb):
    try:
        return Color(rgb=rgb)
    except ValueError:
        return Color(_color_const.WHITE)


def rgba_to_color(rgba):
    return rgb_to_color(rgba[:3])


def rgb_to_hex(rgb):
    return "#" + "".join(
        hex(int_x // 16)[2] + hex(int_x % 16)[2]
        for x in rgb
        for int_x in [int(255 * x)]
    )


def hex_to_rgb(hex_code):
    hex_part = hex_code[1:]
    if len(hex_part) == 3:
        hex_part = "".join([2 * c for c in hex_part])
    return np.array([int(hex_part[i : i + 2], 16) / 255 for i in range(0, 6, 2)])


def invert_color(color):
    return rgb_to_color(1.0 - color_to_rgb(color))


def color_to_int_rgb(color):
    return (255 * color_to_rgb(color)).astype("uint8")


def color_to_int_rgba(color, opacity=1.0):
    alpha = int(255 * opacity)
    return np.array([*color_to_int_rgb(color), alpha])


def random_color():
    return Color(rgb=[np.random.random() for _ in range(3)])



from typing import Tuple, List
import os
from subprocess import Popen, PIPE
from pathlib import Path
from sparrow.multiprocess.config import Config
from sparrow.multiprocess.server import DataManager, register_server_methods
from sparrow.multiprocess.kill_pid import kill_ports


def start_server(port=50001, deque_maxlen=None):
    """
    Create a remote manage server.
    port: 50001
    deque_maxlen: None
    """
    print(f"🚀start success! server port:{port}")
    config = Config(port=port)
    register_server_methods()
    manager = DataManager(address=(config.host, config.port),
                          authkey=config.authkey)
    manager.set_data_deque(maxlen=deque_maxlen)
    manager.get_server().serve_forever()


def run(cmd, **env):
    cmd = cmd.split(' ') if isinstance(cmd, str) else cmd
    p = Popen(cmd, cwd=str(Path(__file__).parent), env={**os.environ, **env})
    return p


def start_server_old():
    server_dir = os.path.dirname(os.path.realpath(__file__))

    p = run(f"python {server_dir}/server.py")
    print('pid:', p.pid)
    print(f"service start at {Config()}")
    p.communicate()
    return p


def stop_server(p: Popen):
    p.kill()


def kill(*ports: int, view=False):
    """
    Kill process by port.
    """
    kill_ports(ports=ports, just_view=view)



from pathlib import Path
from typing import Union
import inspect
import os


def rel_to_abs(rel_path: Union[str, Path], parents=0, return_str=True, strict=False):
    """Return absolute path relative to the called file
    args:
        parent: <int> The number of times `f_back` will be calledd.
    """
    currentframe = inspect.currentframe()
    f = currentframe.f_back
    if parents:
        for _ in range(parents):
            f = f.f_back
    current_path = Path(f.f_code.co_filename).parent
    pathlib_path = current_path / rel_path
    pathlib_path = pathlib_path.resolve(strict=strict)
    if return_str:
        return str(pathlib_path)
    else:
        return pathlib_path


def relp(rel_path: Union[str, Path], parents=0, return_str=True, strict=False):
    return rel_to_abs(rel_path, parents, return_str, strict)


def rel_path_join(*paths: Union[str, Path], return_str=True):
    return rel_to_abs(os.path.join(*paths), parents=1, return_str=return_str)


relp.__doc__ = rel_to_abs.__doc__
